import unittest

from openpectus.lang.exec.units import (
    convert_value_to_unit,
    is_supported_unit,
    get_supported_units,
    get_compatible_unit_names,
    compare_values,
    QUANTITY_UNIT_MAP,
 )


class TestUnits(unittest.TestCase):
    def test_is_supported_unit(self):
        def test_unit(unit: str | None, expect_supported: bool, quantity_name: str = ""):
            if unit is not None:
                test_name = unit if quantity_name == "" else quantity_name + ": " + unit
                with self.subTest(test_name):
                    self.assertEqual(is_supported_unit(unit), expect_supported)

        test_unit("L", True)
        test_unit("%", True)
        test_unit("degC", True)
        test_unit("degF", True)
        test_unit("K", True)
        test_unit("CV", True)
        test_unit("bar", True)
        test_unit("kg/h", True)
        test_unit("AU", True)  # absorbance unit, not 'au'
        test_unit("mS/cm", True)  # milisiemens/cm, electrical conductance
        test_unit("Hz", True)
        test_unit("kHz", True)
        test_unit("m2", True)
        test_unit("kg/L", True)
        test_unit("wt%", True)
        test_unit("LMH", True)
        test_unit("L/m2/h", True)
        test_unit("L/m2/h/bar", True)
        test_unit("°C", True)

        for quantity_name in QUANTITY_UNIT_MAP.keys():
            for unit in QUANTITY_UNIT_MAP[quantity_name]:
                test_unit(unit, True, quantity_name)

        test_unit("", False)
        test_unit("X", False)
        test_unit("ML", False)

    def test_get_compatible_unit_names(self):
        def test_unit(unit: str):
            with self.subTest(unit):
                compatibles = get_compatible_unit_names(unit)
                self.assertIsNotNone(compatibles)
        for unit in get_supported_units():
            if unit is not None:
                test_unit(unit)

    def comp(
            self,
            operator: str,
            value_a: str, unit_a: str | None,
            value_b: str, unit_b: str | None,
            expected_result: bool | None = None,
            expected_value_error_msg: str | None = None):

        s = str(expected_result) if expected_result is not None else expected_value_error_msg or "error"
        sua, sub = unit_a or "", unit_b or ""
        with self.subTest(value_a + sua + " " + operator + " " + value_b + sub + ", " + s):

            result : bool | None = None
            try:
                result = compare_values(operator, value_a, unit_a, value_b, unit_b)
            except ValueError as ve:
                if expected_value_error_msg is None:
                    raise AssertionError(f"compare_value raised ValueError '{ve}' but no error was expected")
                elif str(ve) != expected_value_error_msg:
                    raise AssertionError(f"compare_value raised a ValueError with msg '{str(ve)}' but expected msg " +
                                         f"{expected_value_error_msg}")
            except Exception as ex:
                raise AssertionError("compare_value raised unextected exception: " + str(ex))

            if result is None and expected_result is not None:
                raise AssertionError(f"Expected result={expected_result} but got no result")
            elif result is not None and expected_result is not None:
                if result != expected_result:
                    raise AssertionError(f"Expected result={expected_result} but got result {result}")


    def test_compare_values_numeric_equality(self):
        self.comp("=", "5", None, "5", None, True)
        self.comp("==", "5", None, "5", None, True)
        self.comp("!=", "5", None, "5", None, False)

        self.comp("=", "5", None, "7", None, False)
        self.comp("==", "5", None, "7", None, False)
        self.comp("!=", "5", None, "7", None, True)

    def test_compare_values_numeric_inequality(self):
        self.comp("<", "5", None, "5", None, False)
        self.comp("<=", "5", None, "5", None, True)
        self.comp(">", "5", None, "5", None, False)

        self.comp("<", "5", None, "7", None, True)
        self.comp("<=", "5", None, "7", None, True)
        self.comp(">", "5", None, "7", None, False)

    def test_compare_values_string_equality(self):
        self.comp("=", "foo", None, "foo", None, True)
        self.comp("==", "foo", None, "foo", None, True)
        self.comp("!=", "foo", None, "foo", None, False)

        self.comp("=", "foo", None, "bar", None, False)
        self.comp("==", "foo", None, "bar", None, False)
        self.comp("!=", "foo", None, "bar", None, True)

    def test_compare_values_string_inequality(self):
        self.comp("<", "foo", None, "foo", None, None, "Cannot compare values, first value is missing or not numeric")
        self.comp("<=", "foo", None, "foo", None, None, "Cannot compare values, first value is missing or not numeric")
        self.comp(">", "foo", None, "foo", None, None, "Cannot compare values, first value is missing or not numeric")

        self.comp("<", "5", None, "foo", None, None, "Cannot compare values, second value is missing or not numeric")
        self.comp("<", "", None, "foo", None, None, "Cannot compare values, first value is missing or not numeric")
        self.comp("<", "5", None, "", None, None, "Cannot compare values, second value is missing or not numeric")

    def test_compare_values_invalid_operator(self):
        self.comp("><", "5", None, "7", None, None, "Invalid operator: '><'")

    def test_compare_values_numeric_equality_w_unit(self):
        self.comp("=", "5", "s", "5", "s", True)
        self.comp("=", "5", "s", "5", "s", True)
        self.comp("=", "5", "s", "5.0", "s", True)
        self.comp("==", "5", "s", "5", "s", True)
        self.comp("!=", "5", "s", "5", "s", False)

        self.comp("=", "5", "s", "7", "s", False)
        self.comp("==", "5.0", "s", "7", "s", False)
        self.comp("!=", "5", "s", "7", "s", True)

        self.comp("=", "5", "kg/h", "7", "kg/h", False)
        self.comp("!=", "5", "kg/h", "5", "g/s", True)

        self.comp("=", "60", "L/h", "1", "L/min", True)
        self.comp(">", "61", "L/h", "1", "L/min", True)
        self.comp("<", "59", "L/h", "1", "L/min", True)


    def test_compare_values_numeric_equality_w_unit_error(self):
        self.comp("=", "5", "s", "5", None, None, "Cannot compare values with incompatible units 's' and 'None'")
        self.comp("=", "5", None, "5", "s", None, "Cannot compare values with incompatible units 'None' and 's'")

    def test_compare_values_numeric_equality_w_unit_incompatible(self):
        self.comp("=", "5", "s", "5", "L", None, "Cannot compare values with incompatible units 's' and 'L'")

    def test_compare_values_numeric_inequality_w_unit(self):
        self.comp("<", "5", "s", "5", "s", False)
        self.comp("<", "5", "s", "5", "s", False)

        self.comp("<=", "5", "s", "5", "s", True)
        self.comp(">", "5", "s", "5", "s", False)

        self.comp("<", "5", "s", "7", "s", True)
        self.comp("<=", "5", "s", "7", "s", True)
        self.comp(">", "5", "s", "7", "s", False)

        self.comp(">", "5", "degC", "7", "degC", False)
        self.comp("<", "5", "degC", "7", "degC", True)

    def test_compare_values_compatible_units(self):
        self.comp("=", "60", "s", "1", "min", True)
        self.comp("==", "60", "s", "1", "min", True)
        self.comp("!=", "60", "s", "1", "min", False)

        self.comp("<", "50", "s", "1", "min", True)
        self.comp("<=", "50", "s", "1", "min", True)
        self.comp("<=", "60", "s", "1", "min", True)
        self.comp(">", "70", "s", "1", "min", True)

        self.comp("=", "1", "m2", "100", "dm2", True)
        self.comp("=", "1", "m2", "10000", "cm2", True)
        self.comp("=", "1", "dm2", "100", "cm2", True)

        self.comp("=", "0", "degC", "32", "degF", True)
        self.comp(">", "1", "degC", "32", "degF", True)
        self.comp(">=", "1", "degC", "32", "degF", True)
        self.comp(">", "0", "degC", "33", "degF", False)
        self.comp(">=", "0", "degC", "33", "degF", False)

        self.comp("=", "0", "degC", "273.15", "K", True)
        self.comp(">", "1", "degC", "274", "K", True)
        self.comp(">=", "1", "°C", "274", "K", True)
        self.comp(">", "0", "degC", "274", "K", False)
        self.comp(">=", "0", "degC", "274", "K", False)
        self.comp("=", "13", "degC", "13", "°C", True)

        self.comp("<=", "5", "kg/h", "5000", "g/h", True)
        self.comp(">=", "5", "kg/h", "5000", "g/h", True)
        self.comp("<", "5", "kg/h", "5001", "g/h", True)
        self.comp(">", "5", "kg/h", "4999", "g/h", True)

        self.comp("=", "9", "%", "9", "vol%", True)
        self.comp("=", "9", "%", "9", "wt%", True)
        self.comp("=", "1000", "mAU", "1", "AU", True)
        self.comp("=", "1000", "milliAU", "1", "AU", True)

    def test_compare_multidimensional_pint_units(self):
        # pint treats 'L/d' as 'liter / day' which must be mapped back
        self.comp("=", "48", "L/d", "2", "L/h", True)
        self.comp("<", "48", "L/d", "2.1", "L/h", True)
        self.comp(">", "48.1", "L/d", "2", "L/h", True)
        self.comp("=", "1.4", "LMH", "1.4", "L/m2/h", True)
        self.comp("=", "1.42", "LMH", "1.42", "L/h/m2", True)
        self.comp("=", "1", "kg/L", "1000", "g/L", True)

    def test_troublesome_pint_units(self):
        # pint prefers pascal over Pa so we have defined both. As long as pint
        # is used for comparison via Quantity, it still works.
        self.comp("=", "5", "Pa", "5", "pascal", True)
        self.comp("<", "5", "Pa", "5.1", "pascal", True)
        self.comp(">", "5.1", "Pa", "5", "pascal", True)

    def test_custom_units(self):
        def test(unit: str):
            with self.subTest(unit):
                self.assertTrue(is_supported_unit(unit))

        test("CV")

    def test_compare_custom_units(self):
        self.comp("<", "5", "%", "5.0", "%", False)
        self.comp("==", "5", "%", "5.0", "%", True)

        # TODO: comparison of % to no unit? would seem to be valid 10% == 0.1?

        self.comp("<", "5", "CV", "5.0", "CV", False)
        self.comp("==", "5", "CV", "5.0", "CV", True)

    def test_disallowed_comparison(self):
        with self.assertRaises(ValueError):
            compare_values("=", "9", "vol%", "9", "wt%")
        with self.assertRaises(ValueError):
            compare_values("=", "9", "mol%", "9", "wt%")
        with self.assertRaises(ValueError):
            compare_values("=", "9", "mol%", "9", "vol%")

    def test_convert_value_to_unit(self):
        self.assertAlmostEqual(0.05, convert_value_to_unit(5, "cm", "m"))
        self.assertAlmostEqual(50, convert_value_to_unit(5, "m", "dm"))
        self.assertAlmostEqual(1.0, convert_value_to_unit(60, "L/h", "L/min"))

    def test_convert_value_to_unit_raises_on_incompatible_units(self):
        with self.assertRaises(ValueError):
            convert_value_to_unit(5, "m", "L")



from __future__ import annotations

import logging
from datetime import datetime
from enum import StrEnum, auto
from typing import Literal

import openpectus.aggregator.models as Mdl
from pydantic import BaseModel, ConfigDict
from pydantic.json_schema import SkipJsonSchema

SystemStateEnum = Mdl.SystemStateEnum


class Dto(BaseModel):
    model_config = ConfigDict(from_attributes = True)

    # deliver undefined instead of null for None values. Adapted from:
    # https://github.com/fastapi/fastapi/issues/3314#issuecomment-962932368
    def model_dump(self, *args, **kwargs):
        kwargs.pop('exclude_none', None)
        return super().model_dump(*args, exclude_none=True, **kwargs)


class AuthConfig(Dto):
    use_auth: bool
    authority_url: str | SkipJsonSchema[None] = None
    client_id: str | SkipJsonSchema[None] = None
    well_known_url: str | SkipJsonSchema[None] = None


class ServerErrorResponse(Dto):
    error: bool = True
    message: str


class ServerSuccessResponse(Dto):
    message: str | SkipJsonSchema[None] = None


class ProcessUnitStateEnum(StrEnum):
    """ Represents the state of a process unit. """
    READY = auto()
    ERROR = auto()
    IN_PROGRESS = auto()
    NOT_ONLINE = auto()


class ProcessUnitState:
    class Ready(Dto):
        state: Literal[ProcessUnitStateEnum.READY]

    class Error(Dto):
        state: Literal[ProcessUnitStateEnum.ERROR]

    class InProgress(Dto):
        state: Literal[ProcessUnitStateEnum.IN_PROGRESS]
        progress_pct: int

    class NotOnline(Dto):
        state: Literal[ProcessUnitStateEnum.NOT_ONLINE]
        last_seen_date: datetime


class ProcessDiagram(Dto):
    svg: str


class CommandExample(Dto):
    name: str
    example: str


class UserRole(StrEnum):
    VIEWER = auto()
    ADMIN = auto()


class ControlState(Dto):
    is_running: bool
    is_holding: bool
    is_paused: bool

    @staticmethod
    def default() -> ControlState:
        return ControlState(
            is_running=False,
            is_holding=False,
            is_paused=False)


class ProcessUnit(Dto):
    """Represents a process unit. """
    id: str
    name: str
    state: ProcessUnitState.Ready | ProcessUnitState.Error | ProcessUnitState.InProgress | ProcessUnitState.NotOnline
    location: str | SkipJsonSchema[None] = None
    runtime_msec: int | SkipJsonSchema[None] = None
    current_user_role: UserRole
    uod_author_name: str | SkipJsonSchema[None] = None
    uod_author_email: str | SkipJsonSchema[None] = None
    # users: list[User] ?


class ProcessValueType(StrEnum):
    STRING = auto()
    FLOAT = auto()
    INT = auto()
    CHOICE = auto()
    NONE = auto()


class ProcessValueCommandNumberValue(Dto):
    value: float | int
    value_unit: str | SkipJsonSchema[None] = None
    """ The unit string to display with the value, if any, e.g. 's', 'L/s' or '°C' """
    valid_value_units: list[str] | SkipJsonSchema[None] = None
    """ For values with a unit, provides the list valid alternative units """
    value_type: Literal[ProcessValueType.INT] | Literal[ProcessValueType.FLOAT]
    """ Specifies the type of allowed values. """


class ProcessValueCommandFreeTextValue(Dto):
    value: str
    value_type: Literal[ProcessValueType.STRING]


class ProcessValueCommandChoiceValue(Dto):
    value: str
    value_type: Literal[ProcessValueType.CHOICE]
    options: list[str]


class ProcessValueCommand(Dto):
    command_id: str | SkipJsonSchema[None] = None
    name: str
    command: str
    disabled: bool | SkipJsonSchema[None] = None
    """ Indicates whether the command button should be disabled. """
    value: (ProcessValueCommandNumberValue |
            ProcessValueCommandFreeTextValue |
            ProcessValueCommandChoiceValue |
            SkipJsonSchema[None]) = None


ProcessValueValueType = float | int | str | SkipJsonSchema[None]


def get_ProcessValueType_from_value(value: ProcessValueValueType) -> ProcessValueType:
    if value is None:
        return ProcessValueType.NONE
    if isinstance(value, str):
        return ProcessValueType.STRING
    if isinstance(value, int):
        return ProcessValueType.INT
    if isinstance(value, float):
        return ProcessValueType.FLOAT


class ProcessValue(Dto):
    """ Represents a process value. """
    name: str
    value: ProcessValueValueType = None
    value_formatted: str | SkipJsonSchema[None] = None
    value_unit: str | SkipJsonSchema[None] = None
    """ The unit string to display with the value, if any, e.g. 's', 'L/s' or '°C' """
    value_type: ProcessValueType
    """ Specifies the type of allowed values. """
    commands: list[ProcessValueCommand] | SkipJsonSchema[None] = None
    direction: Mdl.TagDirection

    @staticmethod
    def create(tag: Mdl.TagValue) -> ProcessValue:
        return ProcessValue(
            name=tag.name,
            value=tag.value,
            value_formatted=tag.value_formatted,
            value_type=get_ProcessValueType_from_value(tag.value),
            value_unit=tag.value_unit,
            commands=[],
            direction=tag.direction
        )

    @staticmethod
    def create_w_commands(tag: Mdl.TagValue, commands: list[ProcessValueCommand]) -> ProcessValue:
        pv = ProcessValue.create(tag)
        pv.commands = commands
        return pv


class CommandSource(StrEnum):
    PROCESS_VALUE = auto()
    MANUALLY_ENTERED = auto()
    UNIT_BUTTON = auto()
    """ One of the engine control commands, eg. Start or Pause"""
    METHOD = auto()


class ExecutableCommand(Dto):
    command_id: str | SkipJsonSchema[None] = None
    command: str  # full command string, e.g. "start" or "foo: bar"
    source: CommandSource
    name: str | SkipJsonSchema[None] = None
    value: (ProcessValueCommandNumberValue |
            ProcessValueCommandFreeTextValue |
            ProcessValueCommandChoiceValue |
            SkipJsonSchema[None]) = None


class RunLogLine(Dto):
    id: str
    command: ExecutableCommand
    start: datetime
    end: datetime | SkipJsonSchema[None] = None
    progress: float | SkipJsonSchema[None] = None  # between 0 and 1
    start_values: list[ProcessValue]
    end_values: list[ProcessValue]
    forcible: bool | SkipJsonSchema[None] = None
    cancellable: bool | SkipJsonSchema[None] = None
    forced: bool | SkipJsonSchema[None] = None
    cancelled: bool | SkipJsonSchema[None] = None

    @staticmethod
    def from_model(model: Mdl.RunLogLine) -> RunLogLine:
        return RunLogLine(
            id=model.id,
            command=ExecutableCommand(
                command_id=None,
                command=model.command_name,
                name=None,
                source=CommandSource.METHOD,
                value=None
            ),
            start=datetime.fromtimestamp(model.start),
            end=None if model.end is None else datetime.fromtimestamp(model.end),
            progress=model.progress,
            start_values=list(map(ProcessValue.create, model.start_values)),
            end_values=list(map(ProcessValue.create, model.end_values)),
            forcible=model.forcible,
            forced=model.forced,
            cancellable=model.cancellable,
            cancelled=model.cancelled
        )


class RunLog(Dto):
    lines: list[RunLogLine]

    @staticmethod
    def empty() -> RunLog:
        return RunLog(lines=[])


class MethodLine(Dto):
    id: str
    content: str


class Method(Dto):
    lines: list[MethodLine]

    @staticmethod
    def empty() -> Method:
        return Method(lines=[])


class MethodState(Dto):
    started_line_ids: list[str]
    executed_line_ids: list[str]
    injected_line_ids: list[str]

    @staticmethod
    def empty() -> MethodState:
        return MethodState(started_line_ids=[], executed_line_ids=[], injected_line_ids=[])


class MethodAndState(Dto):
    method: Method
    state: MethodState

    @staticmethod
    def empty() -> MethodAndState:
        return MethodAndState(method=Method.empty(), state=MethodState.empty())


class PlotColorRegion(Dto):
    process_value_name: str
    # Color string compatible with css e.g.:
    # '#aa33bb', 'rgb(0,0,0)', 'rgba(0,0,0,0)', 'red'
    value_color_map: dict[str | int | float, str]


class PlotAxis(Dto):
    label: str
    process_value_names: list[str]
    y_max: int | float
    y_min: int | float
    color: str


class SubPlot(Dto):
    axes: list[PlotAxis]
    ratio: int | float


class PlotConfiguration(Dto):
    process_value_names_to_annotate: list[str]
    color_regions: list[PlotColorRegion]
    sub_plots: list[SubPlot]
    x_axis_process_value_names: list[str]

    @staticmethod
    def empty() -> PlotConfiguration:
        return PlotConfiguration(process_value_names_to_annotate=[],
                                 color_regions=[],
                                 sub_plots=[],
                                 x_axis_process_value_names=[])


# This class originally existed only to workaround the issue that OpenApi spec
# (or Pydantic) cannot express that elements in a list can be
# None/null/undefined.
# Properties on an object can be optional, so we use that via this
# wrapping class to express None values in the PlotLogEntry.values list.
# Feel free to refactor to remove this class if it becomes possible to
# express the above without it, and there is no other need for this class.
class PlotLogEntryValue(Dto):
    value: ProcessValueValueType = None
    tick_time: float


class PlotLogEntry(Dto):
    name: str
    values: list[PlotLogEntryValue]
    value_unit: str | SkipJsonSchema[None] = None
    value_type: ProcessValueType


class PlotLog(Dto):
    entries: dict[str, PlotLogEntry]


class RecentRun(Dto):
    """ Represents a historical run of a process unit. """
    engine_id: str
    run_id: str
    started_date: datetime
    completed_date: datetime
    uod_filename: str
    uod_author_name: str
    uod_author_email: str
    engine_computer_name: str
    engine_version: str
    engine_hardware_str: str
    aggregator_computer_name: str
    aggregator_version: str
    contributors: list[str] = []


class RecentRunMethod(Dto):
    id: str
    run_id: str
    method: Method
    method_state: MethodState


class RecentRunCsv(Dto):
    filename: str
    csv_content: str


class ErrorLogSeverity(StrEnum):
    Warning = auto()
    Error = auto()


class AggregatedErrorLogEntry(Dto):
    message: str
    created_time: datetime
    severity: ErrorLogSeverity
    occurrences: int = 1

    @staticmethod
    def from_model(model: Mdl.AggregatedErrorLogEntry) -> AggregatedErrorLogEntry:
        return AggregatedErrorLogEntry(
            message=model.message,
            created_time=datetime.fromtimestamp(model.created_time),
            severity=ErrorLogSeverity.Error if model.severity == logging.ERROR else ErrorLogSeverity.Warning,
            occurrences=model.occurrences
        )

class AggregatedErrorLog(Dto):
    entries: list[AggregatedErrorLogEntry]

    @staticmethod
    def from_model(model: Mdl.AggregatedErrorLog) -> AggregatedErrorLog:
        return AggregatedErrorLog(entries=[AggregatedErrorLogEntry.from_model(entry) for entry in model.entries])


class BuildInfo(Dto):
    build_number: str
    git_sha: str



import logging
import unittest
from openpectus.engine.hardware import Register, RegisterDirection, HardwareLayerException
from openpectus.engine.labjack_hardware import Labjack_Hardware
import labjack.ljm.constants

logging.basicConfig(format=" %(name)s :: %(levelname)-8s :: %(message)s")
logger = logging.getLogger("openpectus.engine.labjack_hardware.Labjack_Hardware")
logger.setLevel(logging.DEBUG)


SN = labjack.ljm.constants.DEMO_MODE


class TestLabjackHardware(unittest.TestCase):
    def test_can_connect(self):
        hwl = Labjack_Hardware(serial_number=SN)
        try:
            hwl.connect()
            hwl.disconnect()
        except HardwareLayerException:
            self.fail()

    def test_can_read_register(self):
        FT01 = Register("FT01", RegisterDirection.Read, port="AIN1")
        registers = {"FT01": FT01}

        hwl = Labjack_Hardware(serial_number=SN)
        hwl._registers = registers
        with hwl:
            value = hwl.read(FT01)
        self.assertEqual(value, 0.0)

    def test_can_write_register(self):
        PU01 = Register("PU01", RegisterDirection.Both, port="DAC0")
        registers = {"PU01": PU01}

        hwl = Labjack_Hardware(serial_number=SN)
        hwl._registers = registers
        with hwl:
            hwl.write(10.2, PU01)

    def test_can_read_multiple_registers(self):
        registers = {f"Readable {i:02d}": Register(f"{i:02d}",
                                                   RegisterDirection.Read,
                                                   port=f"AIN{i}") for i in range(5)}

        hwl = Labjack_Hardware(serial_number=SN)
        hwl._registers = registers
        hwl.validate_offline()
        with hwl:
            values = hwl.read_batch(list(registers.values()))
        for value in values:
            self.assertEqual(0.0, value)

    def test_can_write_multiple_registers(self):
        registers = {f"Writable {i:02d}": Register(f"{i:02d}",
                     RegisterDirection.Both,
                     port=f"USER_RAM{i}_U16") for i in range(8)}
        values_to_write = [float(i) for i in range(len(registers))]

        hwl = Labjack_Hardware(serial_number=SN)
        hwl._registers = registers
        hwl.validate_offline()
        with hwl:
            hwl.write_batch(values_to_write, list(registers.values()))


if __name__ == "__main__":
    unittest.main()



# Generated from c:/Projects/Novo/Open-Pectus/openpectus/lang/grammar/pcode.g4 by ANTLR 4.13.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,26,377,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,2,7,7,7,2,8,7,8,2,9,7,9,2,10,7,10,2,11,7,11,2,12,7,12,2,13,7,13,
        2,14,7,14,2,15,7,15,2,16,7,16,2,17,7,17,2,18,7,18,2,19,7,19,2,20,
        7,20,2,21,7,21,2,22,7,22,2,23,7,23,2,24,7,24,2,25,7,25,2,26,7,26,
        2,27,7,27,2,28,7,28,2,29,7,29,2,30,7,30,2,31,7,31,2,32,7,32,2,33,
        7,33,1,0,1,0,1,0,5,0,72,8,0,10,0,12,0,75,9,0,1,0,1,0,1,1,5,1,80,
        8,1,10,1,12,1,83,9,1,1,1,1,1,5,1,87,8,1,10,1,12,1,90,9,1,1,1,3,1,
        93,8,1,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,
        2,1,2,3,2,111,8,2,1,3,3,3,114,8,3,1,3,1,3,1,3,5,3,119,8,3,10,3,12,
        3,122,9,3,1,3,3,3,125,8,3,1,3,3,3,128,8,3,1,4,1,4,1,5,3,5,133,8,
        5,1,5,1,5,1,6,3,6,138,8,6,1,6,1,6,1,7,3,7,143,8,7,1,7,1,7,1,7,5,
        7,148,8,7,10,7,12,7,151,9,7,1,7,3,7,154,8,7,1,7,3,7,157,8,7,1,8,
        3,8,160,8,8,1,8,1,8,1,8,5,8,165,8,8,10,8,12,8,168,9,8,1,8,3,8,171,
        8,8,1,8,3,8,174,8,8,1,9,1,9,5,9,178,8,9,10,9,12,9,181,9,9,1,9,1,
        9,5,9,185,8,9,10,9,12,9,188,9,9,1,9,1,9,1,10,1,10,1,11,5,11,195,
        8,11,10,11,12,11,198,9,11,1,11,1,11,1,12,5,12,203,8,12,10,12,12,
        12,206,9,12,1,12,1,12,1,13,3,13,211,8,13,1,13,1,13,1,14,3,14,216,
        8,14,1,14,1,14,1,15,3,15,221,8,15,1,15,1,15,1,16,3,16,226,8,16,1,
        16,1,16,1,16,5,16,231,8,16,10,16,12,16,234,9,16,1,16,3,16,237,8,
        16,1,16,3,16,240,8,16,1,17,3,17,243,8,17,1,17,1,17,1,17,5,17,248,
        8,17,10,17,12,17,251,9,17,1,17,3,17,254,8,17,1,17,3,17,257,8,17,
        1,18,3,18,260,8,18,1,18,1,18,1,18,5,18,265,8,18,10,18,12,18,268,
        9,18,1,18,3,18,271,8,18,1,18,3,18,274,8,18,1,19,5,19,277,8,19,10,
        19,12,19,280,9,19,1,19,1,19,1,20,3,20,285,8,20,1,20,1,20,1,20,5,
        20,290,8,20,10,20,12,20,293,9,20,1,20,3,20,296,8,20,1,21,1,21,1,
        22,1,22,4,22,302,8,22,11,22,12,22,303,1,23,1,23,1,24,1,24,1,24,1,
        25,5,25,312,8,25,10,25,12,25,315,9,25,1,25,1,25,1,26,5,26,320,8,
        26,10,26,12,26,323,9,26,1,27,3,27,326,8,27,1,27,1,27,1,27,5,27,331,
        8,27,10,27,12,27,334,9,27,1,27,3,27,337,8,27,1,28,1,28,1,29,5,29,
        342,8,29,10,29,12,29,345,9,29,1,29,1,29,1,30,1,30,1,31,1,31,5,31,
        353,8,31,10,31,12,31,356,9,31,1,31,3,31,359,8,31,1,32,5,32,362,8,
        32,10,32,12,32,365,9,32,1,32,1,32,1,33,5,33,370,8,33,10,33,12,33,
        373,9,33,1,33,1,33,1,33,8,196,204,278,313,343,354,363,371,0,34,0,
        2,4,6,8,10,12,14,16,18,20,22,24,26,28,30,32,34,36,38,40,42,44,46,
        48,50,52,54,56,58,60,62,64,66,0,4,3,0,15,15,22,23,25,25,2,0,22,23,
        25,25,1,0,25,25,2,0,22,22,25,25,409,0,68,1,0,0,0,2,81,1,0,0,0,4,
        110,1,0,0,0,6,113,1,0,0,0,8,129,1,0,0,0,10,132,1,0,0,0,12,137,1,
        0,0,0,14,142,1,0,0,0,16,159,1,0,0,0,18,175,1,0,0,0,20,191,1,0,0,
        0,22,196,1,0,0,0,24,204,1,0,0,0,26,210,1,0,0,0,28,215,1,0,0,0,30,
        220,1,0,0,0,32,225,1,0,0,0,34,242,1,0,0,0,36,259,1,0,0,0,38,278,
        1,0,0,0,40,284,1,0,0,0,42,297,1,0,0,0,44,299,1,0,0,0,46,305,1,0,
        0,0,48,307,1,0,0,0,50,313,1,0,0,0,52,321,1,0,0,0,54,325,1,0,0,0,
        56,338,1,0,0,0,58,343,1,0,0,0,60,348,1,0,0,0,62,358,1,0,0,0,64,363,
        1,0,0,0,66,371,1,0,0,0,68,73,3,2,1,0,69,70,5,25,0,0,70,72,3,2,1,
        0,71,69,1,0,0,0,72,75,1,0,0,0,73,71,1,0,0,0,73,74,1,0,0,0,74,76,
        1,0,0,0,75,73,1,0,0,0,76,77,5,0,0,1,77,1,1,0,0,0,78,80,5,16,0,0,
        79,78,1,0,0,0,80,83,1,0,0,0,81,79,1,0,0,0,81,82,1,0,0,0,82,84,1,
        0,0,0,83,81,1,0,0,0,84,88,3,4,2,0,85,87,5,16,0,0,86,85,1,0,0,0,87,
        90,1,0,0,0,88,86,1,0,0,0,88,89,1,0,0,0,89,92,1,0,0,0,90,88,1,0,0,
        0,91,93,3,48,24,0,92,91,1,0,0,0,92,93,1,0,0,0,93,3,1,0,0,0,94,111,
        3,6,3,0,95,111,3,10,5,0,96,111,3,12,6,0,97,111,3,14,7,0,98,111,3,
        16,8,0,99,111,3,26,13,0,100,111,3,28,14,0,101,111,3,30,15,0,102,
        111,3,32,16,0,103,111,3,34,17,0,104,111,3,36,18,0,105,111,3,40,20,
        0,106,111,3,54,27,0,107,111,3,48,24,0,108,111,3,52,26,0,109,111,
        3,66,33,0,110,94,1,0,0,0,110,95,1,0,0,0,110,96,1,0,0,0,110,97,1,
        0,0,0,110,98,1,0,0,0,110,99,1,0,0,0,110,100,1,0,0,0,110,101,1,0,
        0,0,110,102,1,0,0,0,110,103,1,0,0,0,110,104,1,0,0,0,110,105,1,0,
        0,0,110,106,1,0,0,0,110,107,1,0,0,0,110,108,1,0,0,0,110,109,1,0,
        0,0,111,5,1,0,0,0,112,114,3,44,22,0,113,112,1,0,0,0,113,114,1,0,
        0,0,114,115,1,0,0,0,115,127,5,9,0,0,116,120,5,23,0,0,117,119,5,16,
        0,0,118,117,1,0,0,0,119,122,1,0,0,0,120,118,1,0,0,0,120,121,1,0,
        0,0,121,123,1,0,0,0,122,120,1,0,0,0,123,125,3,8,4,0,124,116,1,0,
        0,0,124,125,1,0,0,0,125,128,1,0,0,0,126,128,3,64,32,0,127,124,1,
        0,0,0,127,126,1,0,0,0,128,7,1,0,0,0,129,130,3,62,31,0,130,9,1,0,
        0,0,131,133,3,44,22,0,132,131,1,0,0,0,132,133,1,0,0,0,133,134,1,
        0,0,0,134,135,5,10,0,0,135,11,1,0,0,0,136,138,3,44,22,0,137,136,
        1,0,0,0,137,138,1,0,0,0,138,139,1,0,0,0,139,140,5,11,0,0,140,13,
        1,0,0,0,141,143,3,44,22,0,142,141,1,0,0,0,142,143,1,0,0,0,143,144,
        1,0,0,0,144,156,5,1,0,0,145,149,5,23,0,0,146,148,5,16,0,0,147,146,
        1,0,0,0,148,151,1,0,0,0,149,147,1,0,0,0,149,150,1,0,0,0,150,152,
        1,0,0,0,151,149,1,0,0,0,152,154,3,18,9,0,153,145,1,0,0,0,153,154,
        1,0,0,0,154,157,1,0,0,0,155,157,3,64,32,0,156,153,1,0,0,0,156,155,
        1,0,0,0,157,15,1,0,0,0,158,160,3,44,22,0,159,158,1,0,0,0,159,160,
        1,0,0,0,160,161,1,0,0,0,161,173,5,2,0,0,162,166,5,23,0,0,163,165,
        5,16,0,0,164,163,1,0,0,0,165,168,1,0,0,0,166,164,1,0,0,0,166,167,
        1,0,0,0,167,169,1,0,0,0,168,166,1,0,0,0,169,171,3,18,9,0,170,162,
        1,0,0,0,170,171,1,0,0,0,171,174,1,0,0,0,172,174,3,64,32,0,173,170,
        1,0,0,0,173,172,1,0,0,0,174,17,1,0,0,0,175,179,3,22,11,0,176,178,
        5,16,0,0,177,176,1,0,0,0,178,181,1,0,0,0,179,177,1,0,0,0,179,180,
        1,0,0,0,180,182,1,0,0,0,181,179,1,0,0,0,182,186,3,20,10,0,183,185,
        5,16,0,0,184,183,1,0,0,0,185,188,1,0,0,0,186,184,1,0,0,0,186,187,
        1,0,0,0,187,189,1,0,0,0,188,186,1,0,0,0,189,190,3,24,12,0,190,19,
        1,0,0,0,191,192,5,15,0,0,192,21,1,0,0,0,193,195,9,0,0,0,194,193,
        1,0,0,0,195,198,1,0,0,0,196,197,1,0,0,0,196,194,1,0,0,0,197,199,
        1,0,0,0,198,196,1,0,0,0,199,200,8,0,0,0,200,23,1,0,0,0,201,203,9,
        0,0,0,202,201,1,0,0,0,203,206,1,0,0,0,204,205,1,0,0,0,204,202,1,
        0,0,0,205,207,1,0,0,0,206,204,1,0,0,0,207,208,8,0,0,0,208,25,1,0,
        0,0,209,211,3,44,22,0,210,209,1,0,0,0,210,211,1,0,0,0,211,212,1,
        0,0,0,212,213,5,12,0,0,213,27,1,0,0,0,214,216,3,44,22,0,215,214,
        1,0,0,0,215,216,1,0,0,0,216,217,1,0,0,0,217,218,5,7,0,0,218,29,1,
        0,0,0,219,221,3,44,22,0,220,219,1,0,0,0,220,221,1,0,0,0,221,222,
        1,0,0,0,222,223,5,3,0,0,223,31,1,0,0,0,224,226,3,44,22,0,225,224,
        1,0,0,0,225,226,1,0,0,0,226,227,1,0,0,0,227,239,5,4,0,0,228,232,
        5,23,0,0,229,231,5,16,0,0,230,229,1,0,0,0,231,234,1,0,0,0,232,230,
        1,0,0,0,232,233,1,0,0,0,233,235,1,0,0,0,234,232,1,0,0,0,235,237,
        3,38,19,0,236,228,1,0,0,0,236,237,1,0,0,0,237,240,1,0,0,0,238,240,
        3,64,32,0,239,236,1,0,0,0,239,238,1,0,0,0,240,33,1,0,0,0,241,243,
        3,44,22,0,242,241,1,0,0,0,242,243,1,0,0,0,243,244,1,0,0,0,244,256,
        5,5,0,0,245,249,5,23,0,0,246,248,5,16,0,0,247,246,1,0,0,0,248,251,
        1,0,0,0,249,247,1,0,0,0,249,250,1,0,0,0,250,252,1,0,0,0,251,249,
        1,0,0,0,252,254,3,38,19,0,253,245,1,0,0,0,253,254,1,0,0,0,254,257,
        1,0,0,0,255,257,3,64,32,0,256,253,1,0,0,0,256,255,1,0,0,0,257,35,
        1,0,0,0,258,260,3,44,22,0,259,258,1,0,0,0,259,260,1,0,0,0,260,261,
        1,0,0,0,261,273,5,6,0,0,262,266,5,23,0,0,263,265,5,16,0,0,264,263,
        1,0,0,0,265,268,1,0,0,0,266,264,1,0,0,0,266,267,1,0,0,0,267,269,
        1,0,0,0,268,266,1,0,0,0,269,271,3,38,19,0,270,262,1,0,0,0,270,271,
        1,0,0,0,271,274,1,0,0,0,272,274,3,64,32,0,273,270,1,0,0,0,273,272,
        1,0,0,0,274,37,1,0,0,0,275,277,9,0,0,0,276,275,1,0,0,0,277,280,1,
        0,0,0,278,279,1,0,0,0,278,276,1,0,0,0,279,281,1,0,0,0,280,278,1,
        0,0,0,281,282,8,1,0,0,282,39,1,0,0,0,283,285,3,44,22,0,284,283,1,
        0,0,0,284,285,1,0,0,0,285,286,1,0,0,0,286,287,5,8,0,0,287,291,5,
        23,0,0,288,290,5,16,0,0,289,288,1,0,0,0,290,293,1,0,0,0,291,289,
        1,0,0,0,291,292,1,0,0,0,292,295,1,0,0,0,293,291,1,0,0,0,294,296,
        3,42,21,0,295,294,1,0,0,0,295,296,1,0,0,0,296,41,1,0,0,0,297,298,
        3,62,31,0,298,43,1,0,0,0,299,301,3,46,23,0,300,302,5,16,0,0,301,
        300,1,0,0,0,302,303,1,0,0,0,303,301,1,0,0,0,303,304,1,0,0,0,304,
        45,1,0,0,0,305,306,5,14,0,0,306,47,1,0,0,0,307,308,5,22,0,0,308,
        309,3,50,25,0,309,49,1,0,0,0,310,312,9,0,0,0,311,310,1,0,0,0,312,
        315,1,0,0,0,313,314,1,0,0,0,313,311,1,0,0,0,314,316,1,0,0,0,315,
        313,1,0,0,0,316,317,8,2,0,0,317,51,1,0,0,0,318,320,5,16,0,0,319,
        318,1,0,0,0,320,323,1,0,0,0,321,319,1,0,0,0,321,322,1,0,0,0,322,
        53,1,0,0,0,323,321,1,0,0,0,324,326,3,44,22,0,325,324,1,0,0,0,325,
        326,1,0,0,0,326,327,1,0,0,0,327,336,3,56,28,0,328,332,5,23,0,0,329,
        331,5,16,0,0,330,329,1,0,0,0,331,334,1,0,0,0,332,330,1,0,0,0,332,
        333,1,0,0,0,333,335,1,0,0,0,334,332,1,0,0,0,335,337,3,58,29,0,336,
        328,1,0,0,0,336,337,1,0,0,0,337,55,1,0,0,0,338,339,3,60,30,0,339,
        57,1,0,0,0,340,342,9,0,0,0,341,340,1,0,0,0,342,345,1,0,0,0,343,344,
        1,0,0,0,343,341,1,0,0,0,344,346,1,0,0,0,345,343,1,0,0,0,346,347,
        8,3,0,0,347,59,1,0,0,0,348,349,5,13,0,0,349,61,1,0,0,0,350,359,5,
        13,0,0,351,353,9,0,0,0,352,351,1,0,0,0,353,356,1,0,0,0,354,355,1,
        0,0,0,354,352,1,0,0,0,355,357,1,0,0,0,356,354,1,0,0,0,357,359,8,
        3,0,0,358,350,1,0,0,0,358,354,1,0,0,0,359,63,1,0,0,0,360,362,9,0,
        0,0,361,360,1,0,0,0,362,365,1,0,0,0,363,364,1,0,0,0,363,361,1,0,
        0,0,364,366,1,0,0,0,365,363,1,0,0,0,366,367,8,3,0,0,367,65,1,0,0,
        0,368,370,9,0,0,0,369,368,1,0,0,0,370,373,1,0,0,0,371,372,1,0,0,
        0,371,369,1,0,0,0,372,374,1,0,0,0,373,371,1,0,0,0,374,375,8,3,0,
        0,375,67,1,0,0,0,53,73,81,88,92,110,113,120,124,127,132,137,142,
        149,153,156,159,166,170,173,179,186,196,204,210,215,220,225,232,
        236,239,242,249,253,256,259,266,270,273,278,284,291,295,303,313,
        321,325,332,336,343,354,358,363,371
    ]

class pcodeParser ( Parser ):

    grammarFileName = "pcode.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'Watch'", "'Alarm'", "'Stop'", "'Pause'", 
                     "'Hold'", "'Wait'", "'Restart'", "'Mark'", "'Block'", 
                     "'End block'", "'End blocks'", "'Increment run counter'", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "'_'", "'.'", "','", "' '", "'\\t'", "'#'", "':'", 
                     "'-'" ]

    symbolicNames = [ "<INVALID>", "WATCH", "ALARM", "STOP", "PAUSE", "HOLD", 
                      "WAIT", "RESTART", "MARK", "BLOCK", "END_BLOCK", "END_BLOCKS", 
                      "INCREMENT_RC", "IDENTIFIER", "POSITIVE_FLOAT", "COMPARE_OP", 
                      "WHITESPACE", "UNDERSCORE", "PERIOD", "COMMA", "SPACE", 
                      "TAB", "HASH", "COLON", "MINUS", "NEWLINE", "ANY" ]

    RULE_program = 0
    RULE_instruction_line = 1
    RULE_instruction = 2
    RULE_block = 3
    RULE_block_name = 4
    RULE_end_block = 5
    RULE_end_blocks = 6
    RULE_watch = 7
    RULE_alarm = 8
    RULE_condition = 9
    RULE_compare_op = 10
    RULE_condition_lhs = 11
    RULE_condition_rhs = 12
    RULE_increment_rc = 13
    RULE_restart = 14
    RULE_stop = 15
    RULE_pause = 16
    RULE_hold = 17
    RULE_wait = 18
    RULE_duration = 19
    RULE_mark = 20
    RULE_mark_name = 21
    RULE_time = 22
    RULE_timeexp = 23
    RULE_comment = 24
    RULE_comment_text = 25
    RULE_blank = 26
    RULE_command = 27
    RULE_command_name = 28
    RULE_command_args = 29
    RULE_identifier = 30
    RULE_identifier_ext = 31
    RULE_inst_error = 32
    RULE_error = 33

    ruleNames =  [ "program", "instruction_line", "instruction", "block", 
                   "block_name", "end_block", "end_blocks", "watch", "alarm", 
                   "condition", "compare_op", "condition_lhs", "condition_rhs", 
                   "increment_rc", "restart", "stop", "pause", "hold", "wait", 
                   "duration", "mark", "mark_name", "time", "timeexp", "comment", 
                   "comment_text", "blank", "command", "command_name", "command_args", 
                   "identifier", "identifier_ext", "inst_error", "error" ]

    EOF = Token.EOF
    WATCH=1
    ALARM=2
    STOP=3
    PAUSE=4
    HOLD=5
    WAIT=6
    RESTART=7
    MARK=8
    BLOCK=9
    END_BLOCK=10
    END_BLOCKS=11
    INCREMENT_RC=12
    IDENTIFIER=13
    POSITIVE_FLOAT=14
    COMPARE_OP=15
    WHITESPACE=16
    UNDERSCORE=17
    PERIOD=18
    COMMA=19
    SPACE=20
    TAB=21
    HASH=22
    COLON=23
    MINUS=24
    NEWLINE=25
    ANY=26

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class ProgramContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def instruction_line(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(pcodeParser.Instruction_lineContext)
            else:
                return self.getTypedRuleContext(pcodeParser.Instruction_lineContext,i)


        def EOF(self):
            return self.getToken(pcodeParser.EOF, 0)

        def NEWLINE(self, i:int=None):
            if i is None:
                return self.getTokens(pcodeParser.NEWLINE)
            else:
                return self.getToken(pcodeParser.NEWLINE, i)

        def getRuleIndex(self):
            return pcodeParser.RULE_program

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterProgram" ):
                listener.enterProgram(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitProgram" ):
                listener.exitProgram(self)




    def program(self):

        localctx = pcodeParser.ProgramContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_program)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 68
            self.instruction_line()
            self.state = 73
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==25:
                self.state = 69
                self.match(pcodeParser.NEWLINE)
                self.state = 70
                self.instruction_line()
                self.state = 75
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 76
            self.match(pcodeParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Instruction_lineContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def instruction(self):
            return self.getTypedRuleContext(pcodeParser.InstructionContext,0)


        def WHITESPACE(self, i:int=None):
            if i is None:
                return self.getTokens(pcodeParser.WHITESPACE)
            else:
                return self.getToken(pcodeParser.WHITESPACE, i)

        def comment(self):
            return self.getTypedRuleContext(pcodeParser.CommentContext,0)


        def getRuleIndex(self):
            return pcodeParser.RULE_instruction_line

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterInstruction_line" ):
                listener.enterInstruction_line(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitInstruction_line" ):
                listener.exitInstruction_line(self)




    def instruction_line(self):

        localctx = pcodeParser.Instruction_lineContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_instruction_line)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 81
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,1,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    self.state = 78
                    self.match(pcodeParser.WHITESPACE) 
                self.state = 83
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,1,self._ctx)

            self.state = 84
            self.instruction()
            self.state = 88
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==16:
                self.state = 85
                self.match(pcodeParser.WHITESPACE)
                self.state = 90
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 92
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==22:
                self.state = 91
                self.comment()


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class InstructionContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def block(self):
            return self.getTypedRuleContext(pcodeParser.BlockContext,0)


        def end_block(self):
            return self.getTypedRuleContext(pcodeParser.End_blockContext,0)


        def end_blocks(self):
            return self.getTypedRuleContext(pcodeParser.End_blocksContext,0)


        def watch(self):
            return self.getTypedRuleContext(pcodeParser.WatchContext,0)


        def alarm(self):
            return self.getTypedRuleContext(pcodeParser.AlarmContext,0)


        def increment_rc(self):
            return self.getTypedRuleContext(pcodeParser.Increment_rcContext,0)


        def restart(self):
            return self.getTypedRuleContext(pcodeParser.RestartContext,0)


        def stop(self):
            return self.getTypedRuleContext(pcodeParser.StopContext,0)


        def pause(self):
            return self.getTypedRuleContext(pcodeParser.PauseContext,0)


        def hold(self):
            return self.getTypedRuleContext(pcodeParser.HoldContext,0)


        def wait(self):
            return self.getTypedRuleContext(pcodeParser.WaitContext,0)


        def mark(self):
            return self.getTypedRuleContext(pcodeParser.MarkContext,0)


        def command(self):
            return self.getTypedRuleContext(pcodeParser.CommandContext,0)


        def comment(self):
            return self.getTypedRuleContext(pcodeParser.CommentContext,0)


        def blank(self):
            return self.getTypedRuleContext(pcodeParser.BlankContext,0)


        def error(self):
            return self.getTypedRuleContext(pcodeParser.ErrorContext,0)


        def getRuleIndex(self):
            return pcodeParser.RULE_instruction

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterInstruction" ):
                listener.enterInstruction(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitInstruction" ):
                listener.exitInstruction(self)




    def instruction(self):

        localctx = pcodeParser.InstructionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_instruction)
        try:
            self.state = 110
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,4,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 94
                self.block()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 95
                self.end_block()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 96
                self.end_blocks()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 97
                self.watch()
                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 98
                self.alarm()
                pass

            elif la_ == 6:
                self.enterOuterAlt(localctx, 6)
                self.state = 99
                self.increment_rc()
                pass

            elif la_ == 7:
                self.enterOuterAlt(localctx, 7)
                self.state = 100
                self.restart()
                pass

            elif la_ == 8:
                self.enterOuterAlt(localctx, 8)
                self.state = 101
                self.stop()
                pass

            elif la_ == 9:
                self.enterOuterAlt(localctx, 9)
                self.state = 102
                self.pause()
                pass

            elif la_ == 10:
                self.enterOuterAlt(localctx, 10)
                self.state = 103
                self.hold()
                pass

            elif la_ == 11:
                self.enterOuterAlt(localctx, 11)
                self.state = 104
                self.wait()
                pass

            elif la_ == 12:
                self.enterOuterAlt(localctx, 12)
                self.state = 105
                self.mark()
                pass

            elif la_ == 13:
                self.enterOuterAlt(localctx, 13)
                self.state = 106
                self.command()
                pass

            elif la_ == 14:
                self.enterOuterAlt(localctx, 14)
                self.state = 107
                self.comment()
                pass

            elif la_ == 15:
                self.enterOuterAlt(localctx, 15)
                self.state = 108
                self.blank()
                pass

            elif la_ == 16:
                self.enterOuterAlt(localctx, 16)
                self.state = 109
                self.error()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class BlockContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def BLOCK(self):
            return self.getToken(pcodeParser.BLOCK, 0)

        def inst_error(self):
            return self.getTypedRuleContext(pcodeParser.Inst_errorContext,0)


        def time(self):
            return self.getTypedRuleContext(pcodeParser.TimeContext,0)


        def COLON(self):
            return self.getToken(pcodeParser.COLON, 0)

        def block_name(self):
            return self.getTypedRuleContext(pcodeParser.Block_nameContext,0)


        def WHITESPACE(self, i:int=None):
            if i is None:
                return self.getTokens(pcodeParser.WHITESPACE)
            else:
                return self.getToken(pcodeParser.WHITESPACE, i)

        def getRuleIndex(self):
            return pcodeParser.RULE_block

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBlock" ):
                listener.enterBlock(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBlock" ):
                listener.exitBlock(self)




    def block(self):

        localctx = pcodeParser.BlockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_block)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 113
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==14:
                self.state = 112
                self.time()


            self.state = 115
            self.match(pcodeParser.BLOCK)
            self.state = 127
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,8,self._ctx)
            if la_ == 1:
                self.state = 124
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==23:
                    self.state = 116
                    self.match(pcodeParser.COLON)
                    self.state = 120
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,6,self._ctx)
                    while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                        if _alt==1:
                            self.state = 117
                            self.match(pcodeParser.WHITESPACE) 
                        self.state = 122
                        self._errHandler.sync(self)
                        _alt = self._interp.adaptivePredict(self._input,6,self._ctx)

                    self.state = 123
                    self.block_name()


                pass

            elif la_ == 2:
                self.state = 126
                self.inst_error()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Block_nameContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def identifier_ext(self):
            return self.getTypedRuleContext(pcodeParser.Identifier_extContext,0)


        def getRuleIndex(self):
            return pcodeParser.RULE_block_name

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBlock_name" ):
                listener.enterBlock_name(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBlock_name" ):
                listener.exitBlock_name(self)




    def block_name(self):

        localctx = pcodeParser.Block_nameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_block_name)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 129
            self.identifier_ext()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class End_blockContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def END_BLOCK(self):
            return self.getToken(pcodeParser.END_BLOCK, 0)

        def time(self):
            return self.getTypedRuleContext(pcodeParser.TimeContext,0)


        def getRuleIndex(self):
            return pcodeParser.RULE_end_block

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterEnd_block" ):
                listener.enterEnd_block(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitEnd_block" ):
                listener.exitEnd_block(self)




    def end_block(self):

        localctx = pcodeParser.End_blockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_end_block)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 132
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==14:
                self.state = 131
                self.time()


            self.state = 134
            self.match(pcodeParser.END_BLOCK)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class End_blocksContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def END_BLOCKS(self):
            return self.getToken(pcodeParser.END_BLOCKS, 0)

        def time(self):
            return self.getTypedRuleContext(pcodeParser.TimeContext,0)


        def getRuleIndex(self):
            return pcodeParser.RULE_end_blocks

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterEnd_blocks" ):
                listener.enterEnd_blocks(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitEnd_blocks" ):
                listener.exitEnd_blocks(self)




    def end_blocks(self):

        localctx = pcodeParser.End_blocksContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_end_blocks)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 137
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==14:
                self.state = 136
                self.time()


            self.state = 139
            self.match(pcodeParser.END_BLOCKS)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class WatchContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def WATCH(self):
            return self.getToken(pcodeParser.WATCH, 0)

        def inst_error(self):
            return self.getTypedRuleContext(pcodeParser.Inst_errorContext,0)


        def time(self):
            return self.getTypedRuleContext(pcodeParser.TimeContext,0)


        def COLON(self):
            return self.getToken(pcodeParser.COLON, 0)

        def condition(self):
            return self.getTypedRuleContext(pcodeParser.ConditionContext,0)


        def WHITESPACE(self, i:int=None):
            if i is None:
                return self.getTokens(pcodeParser.WHITESPACE)
            else:
                return self.getToken(pcodeParser.WHITESPACE, i)

        def getRuleIndex(self):
            return pcodeParser.RULE_watch

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterWatch" ):
                listener.enterWatch(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitWatch" ):
                listener.exitWatch(self)




    def watch(self):

        localctx = pcodeParser.WatchContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_watch)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 142
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==14:
                self.state = 141
                self.time()


            self.state = 144
            self.match(pcodeParser.WATCH)
            self.state = 156
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,14,self._ctx)
            if la_ == 1:
                self.state = 153
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==23:
                    self.state = 145
                    self.match(pcodeParser.COLON)
                    self.state = 149
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,12,self._ctx)
                    while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                        if _alt==1:
                            self.state = 146
                            self.match(pcodeParser.WHITESPACE) 
                        self.state = 151
                        self._errHandler.sync(self)
                        _alt = self._interp.adaptivePredict(self._input,12,self._ctx)

                    self.state = 152
                    self.condition()


                pass

            elif la_ == 2:
                self.state = 155
                self.inst_error()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AlarmContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ALARM(self):
            return self.getToken(pcodeParser.ALARM, 0)

        def inst_error(self):
            return self.getTypedRuleContext(pcodeParser.Inst_errorContext,0)


        def time(self):
            return self.getTypedRuleContext(pcodeParser.TimeContext,0)


        def COLON(self):
            return self.getToken(pcodeParser.COLON, 0)

        def condition(self):
            return self.getTypedRuleContext(pcodeParser.ConditionContext,0)


        def WHITESPACE(self, i:int=None):
            if i is None:
                return self.getTokens(pcodeParser.WHITESPACE)
            else:
                return self.getToken(pcodeParser.WHITESPACE, i)

        def getRuleIndex(self):
            return pcodeParser.RULE_alarm

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAlarm" ):
                listener.enterAlarm(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAlarm" ):
                listener.exitAlarm(self)




    def alarm(self):

        localctx = pcodeParser.AlarmContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_alarm)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 159
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==14:
                self.state = 158
                self.time()


            self.state = 161
            self.match(pcodeParser.ALARM)
            self.state = 173
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,18,self._ctx)
            if la_ == 1:
                self.state = 170
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==23:
                    self.state = 162
                    self.match(pcodeParser.COLON)
                    self.state = 166
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,16,self._ctx)
                    while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                        if _alt==1:
                            self.state = 163
                            self.match(pcodeParser.WHITESPACE) 
                        self.state = 168
                        self._errHandler.sync(self)
                        _alt = self._interp.adaptivePredict(self._input,16,self._ctx)

                    self.state = 169
                    self.condition()


                pass

            elif la_ == 2:
                self.state = 172
                self.inst_error()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ConditionContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def condition_lhs(self):
            return self.getTypedRuleContext(pcodeParser.Condition_lhsContext,0)


        def compare_op(self):
            return self.getTypedRuleContext(pcodeParser.Compare_opContext,0)


        def condition_rhs(self):
            return self.getTypedRuleContext(pcodeParser.Condition_rhsContext,0)


        def WHITESPACE(self, i:int=None):
            if i is None:
                return self.getTokens(pcodeParser.WHITESPACE)
            else:
                return self.getToken(pcodeParser.WHITESPACE, i)

        def getRuleIndex(self):
            return pcodeParser.RULE_condition

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCondition" ):
                listener.enterCondition(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCondition" ):
                listener.exitCondition(self)




    def condition(self):

        localctx = pcodeParser.ConditionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_condition)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 175
            self.condition_lhs()
            self.state = 179
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==16:
                self.state = 176
                self.match(pcodeParser.WHITESPACE)
                self.state = 181
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 182
            self.compare_op()
            self.state = 186
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,20,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    self.state = 183
                    self.match(pcodeParser.WHITESPACE) 
                self.state = 188
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,20,self._ctx)

            self.state = 189
            self.condition_rhs()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Compare_opContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def COMPARE_OP(self):
            return self.getToken(pcodeParser.COMPARE_OP, 0)

        def getRuleIndex(self):
            return pcodeParser.RULE_compare_op

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCompare_op" ):
                listener.enterCompare_op(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCompare_op" ):
                listener.exitCompare_op(self)




    def compare_op(self):

        localctx = pcodeParser.Compare_opContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_compare_op)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 191
            self.match(pcodeParser.COMPARE_OP)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Condition_lhsContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NEWLINE(self):
            return self.getToken(pcodeParser.NEWLINE, 0)

        def HASH(self):
            return self.getToken(pcodeParser.HASH, 0)

        def COMPARE_OP(self):
            return self.getToken(pcodeParser.COMPARE_OP, 0)

        def COLON(self):
            return self.getToken(pcodeParser.COLON, 0)

        def getRuleIndex(self):
            return pcodeParser.RULE_condition_lhs

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCondition_lhs" ):
                listener.enterCondition_lhs(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCondition_lhs" ):
                listener.exitCondition_lhs(self)




    def condition_lhs(self):

        localctx = pcodeParser.Condition_lhsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_condition_lhs)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 196
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,21,self._ctx)
            while _alt!=1 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1+1:
                    self.state = 193
                    self.matchWildcard() 
                self.state = 198
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,21,self._ctx)

            self.state = 199
            _la = self._input.LA(1)
            if _la <= 0 or (((_la) & ~0x3f) == 0 and ((1 << _la) & 46170112) != 0):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Condition_rhsContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NEWLINE(self):
            return self.getToken(pcodeParser.NEWLINE, 0)

        def HASH(self):
            return self.getToken(pcodeParser.HASH, 0)

        def COMPARE_OP(self):
            return self.getToken(pcodeParser.COMPARE_OP, 0)

        def COLON(self):
            return self.getToken(pcodeParser.COLON, 0)

        def getRuleIndex(self):
            return pcodeParser.RULE_condition_rhs

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCondition_rhs" ):
                listener.enterCondition_rhs(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCondition_rhs" ):
                listener.exitCondition_rhs(self)




    def condition_rhs(self):

        localctx = pcodeParser.Condition_rhsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_condition_rhs)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 204
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,22,self._ctx)
            while _alt!=1 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1+1:
                    self.state = 201
                    self.matchWildcard() 
                self.state = 206
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,22,self._ctx)

            self.state = 207
            _la = self._input.LA(1)
            if _la <= 0 or (((_la) & ~0x3f) == 0 and ((1 << _la) & 46170112) != 0):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Increment_rcContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def INCREMENT_RC(self):
            return self.getToken(pcodeParser.INCREMENT_RC, 0)

        def time(self):
            return self.getTypedRuleContext(pcodeParser.TimeContext,0)


        def getRuleIndex(self):
            return pcodeParser.RULE_increment_rc

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterIncrement_rc" ):
                listener.enterIncrement_rc(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitIncrement_rc" ):
                listener.exitIncrement_rc(self)




    def increment_rc(self):

        localctx = pcodeParser.Increment_rcContext(self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_increment_rc)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 210
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==14:
                self.state = 209
                self.time()


            self.state = 212
            self.match(pcodeParser.INCREMENT_RC)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class RestartContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def RESTART(self):
            return self.getToken(pcodeParser.RESTART, 0)

        def time(self):
            return self.getTypedRuleContext(pcodeParser.TimeContext,0)


        def getRuleIndex(self):
            return pcodeParser.RULE_restart

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterRestart" ):
                listener.enterRestart(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitRestart" ):
                listener.exitRestart(self)




    def restart(self):

        localctx = pcodeParser.RestartContext(self, self._ctx, self.state)
        self.enterRule(localctx, 28, self.RULE_restart)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 215
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==14:
                self.state = 214
                self.time()


            self.state = 217
            self.match(pcodeParser.RESTART)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class StopContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STOP(self):
            return self.getToken(pcodeParser.STOP, 0)

        def time(self):
            return self.getTypedRuleContext(pcodeParser.TimeContext,0)


        def getRuleIndex(self):
            return pcodeParser.RULE_stop

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterStop" ):
                listener.enterStop(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitStop" ):
                listener.exitStop(self)




    def stop(self):

        localctx = pcodeParser.StopContext(self, self._ctx, self.state)
        self.enterRule(localctx, 30, self.RULE_stop)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 220
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==14:
                self.state = 219
                self.time()


            self.state = 222
            self.match(pcodeParser.STOP)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PauseContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def PAUSE(self):
            return self.getToken(pcodeParser.PAUSE, 0)

        def inst_error(self):
            return self.getTypedRuleContext(pcodeParser.Inst_errorContext,0)


        def time(self):
            return self.getTypedRuleContext(pcodeParser.TimeContext,0)


        def COLON(self):
            return self.getToken(pcodeParser.COLON, 0)

        def duration(self):
            return self.getTypedRuleContext(pcodeParser.DurationContext,0)


        def WHITESPACE(self, i:int=None):
            if i is None:
                return self.getTokens(pcodeParser.WHITESPACE)
            else:
                return self.getToken(pcodeParser.WHITESPACE, i)

        def getRuleIndex(self):
            return pcodeParser.RULE_pause

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPause" ):
                listener.enterPause(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPause" ):
                listener.exitPause(self)




    def pause(self):

        localctx = pcodeParser.PauseContext(self, self._ctx, self.state)
        self.enterRule(localctx, 32, self.RULE_pause)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 225
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==14:
                self.state = 224
                self.time()


            self.state = 227
            self.match(pcodeParser.PAUSE)
            self.state = 239
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,29,self._ctx)
            if la_ == 1:
                self.state = 236
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==23:
                    self.state = 228
                    self.match(pcodeParser.COLON)
                    self.state = 232
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,27,self._ctx)
                    while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                        if _alt==1:
                            self.state = 229
                            self.match(pcodeParser.WHITESPACE) 
                        self.state = 234
                        self._errHandler.sync(self)
                        _alt = self._interp.adaptivePredict(self._input,27,self._ctx)

                    self.state = 235
                    self.duration()


                pass

            elif la_ == 2:
                self.state = 238
                self.inst_error()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class HoldContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def HOLD(self):
            return self.getToken(pcodeParser.HOLD, 0)

        def inst_error(self):
            return self.getTypedRuleContext(pcodeParser.Inst_errorContext,0)


        def time(self):
            return self.getTypedRuleContext(pcodeParser.TimeContext,0)


        def COLON(self):
            return self.getToken(pcodeParser.COLON, 0)

        def duration(self):
            return self.getTypedRuleContext(pcodeParser.DurationContext,0)


        def WHITESPACE(self, i:int=None):
            if i is None:
                return self.getTokens(pcodeParser.WHITESPACE)
            else:
                return self.getToken(pcodeParser.WHITESPACE, i)

        def getRuleIndex(self):
            return pcodeParser.RULE_hold

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterHold" ):
                listener.enterHold(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitHold" ):
                listener.exitHold(self)




    def hold(self):

        localctx = pcodeParser.HoldContext(self, self._ctx, self.state)
        self.enterRule(localctx, 34, self.RULE_hold)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 242
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==14:
                self.state = 241
                self.time()


            self.state = 244
            self.match(pcodeParser.HOLD)
            self.state = 256
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,33,self._ctx)
            if la_ == 1:
                self.state = 253
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==23:
                    self.state = 245
                    self.match(pcodeParser.COLON)
                    self.state = 249
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,31,self._ctx)
                    while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                        if _alt==1:
                            self.state = 246
                            self.match(pcodeParser.WHITESPACE) 
                        self.state = 251
                        self._errHandler.sync(self)
                        _alt = self._interp.adaptivePredict(self._input,31,self._ctx)

                    self.state = 252
                    self.duration()


                pass

            elif la_ == 2:
                self.state = 255
                self.inst_error()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class WaitContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def WAIT(self):
            return self.getToken(pcodeParser.WAIT, 0)

        def inst_error(self):
            return self.getTypedRuleContext(pcodeParser.Inst_errorContext,0)


        def time(self):
            return self.getTypedRuleContext(pcodeParser.TimeContext,0)


        def COLON(self):
            return self.getToken(pcodeParser.COLON, 0)

        def duration(self):
            return self.getTypedRuleContext(pcodeParser.DurationContext,0)


        def WHITESPACE(self, i:int=None):
            if i is None:
                return self.getTokens(pcodeParser.WHITESPACE)
            else:
                return self.getToken(pcodeParser.WHITESPACE, i)

        def getRuleIndex(self):
            return pcodeParser.RULE_wait

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterWait" ):
                listener.enterWait(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitWait" ):
                listener.exitWait(self)




    def wait(self):

        localctx = pcodeParser.WaitContext(self, self._ctx, self.state)
        self.enterRule(localctx, 36, self.RULE_wait)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 259
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==14:
                self.state = 258
                self.time()


            self.state = 261
            self.match(pcodeParser.WAIT)
            self.state = 273
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,37,self._ctx)
            if la_ == 1:
                self.state = 270
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==23:
                    self.state = 262
                    self.match(pcodeParser.COLON)
                    self.state = 266
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,35,self._ctx)
                    while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                        if _alt==1:
                            self.state = 263
                            self.match(pcodeParser.WHITESPACE) 
                        self.state = 268
                        self._errHandler.sync(self)
                        _alt = self._interp.adaptivePredict(self._input,35,self._ctx)

                    self.state = 269
                    self.duration()


                pass

            elif la_ == 2:
                self.state = 272
                self.inst_error()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class DurationContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NEWLINE(self):
            return self.getToken(pcodeParser.NEWLINE, 0)

        def HASH(self):
            return self.getToken(pcodeParser.HASH, 0)

        def COLON(self):
            return self.getToken(pcodeParser.COLON, 0)

        def getRuleIndex(self):
            return pcodeParser.RULE_duration

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDuration" ):
                listener.enterDuration(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDuration" ):
                listener.exitDuration(self)




    def duration(self):

        localctx = pcodeParser.DurationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 38, self.RULE_duration)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 278
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,38,self._ctx)
            while _alt!=1 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1+1:
                    self.state = 275
                    self.matchWildcard() 
                self.state = 280
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,38,self._ctx)

            self.state = 281
            _la = self._input.LA(1)
            if _la <= 0 or (((_la) & ~0x3f) == 0 and ((1 << _la) & 46137344) != 0):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class MarkContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def MARK(self):
            return self.getToken(pcodeParser.MARK, 0)

        def COLON(self):
            return self.getToken(pcodeParser.COLON, 0)

        def time(self):
            return self.getTypedRuleContext(pcodeParser.TimeContext,0)


        def WHITESPACE(self, i:int=None):
            if i is None:
                return self.getTokens(pcodeParser.WHITESPACE)
            else:
                return self.getToken(pcodeParser.WHITESPACE, i)

        def mark_name(self):
            return self.getTypedRuleContext(pcodeParser.Mark_nameContext,0)


        def getRuleIndex(self):
            return pcodeParser.RULE_mark

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterMark" ):
                listener.enterMark(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitMark" ):
                listener.exitMark(self)




    def mark(self):

        localctx = pcodeParser.MarkContext(self, self._ctx, self.state)
        self.enterRule(localctx, 40, self.RULE_mark)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 284
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==14:
                self.state = 283
                self.time()


            self.state = 286
            self.match(pcodeParser.MARK)
            self.state = 287
            self.match(pcodeParser.COLON)
            self.state = 291
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,40,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    self.state = 288
                    self.match(pcodeParser.WHITESPACE) 
                self.state = 293
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,40,self._ctx)

            self.state = 295
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,41,self._ctx)
            if la_ == 1:
                self.state = 294
                self.mark_name()


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Mark_nameContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def identifier_ext(self):
            return self.getTypedRuleContext(pcodeParser.Identifier_extContext,0)


        def getRuleIndex(self):
            return pcodeParser.RULE_mark_name

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterMark_name" ):
                listener.enterMark_name(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitMark_name" ):
                listener.exitMark_name(self)




    def mark_name(self):

        localctx = pcodeParser.Mark_nameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 42, self.RULE_mark_name)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 297
            self.identifier_ext()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class TimeContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def timeexp(self):
            return self.getTypedRuleContext(pcodeParser.TimeexpContext,0)


        def WHITESPACE(self, i:int=None):
            if i is None:
                return self.getTokens(pcodeParser.WHITESPACE)
            else:
                return self.getToken(pcodeParser.WHITESPACE, i)

        def getRuleIndex(self):
            return pcodeParser.RULE_time

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterTime" ):
                listener.enterTime(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitTime" ):
                listener.exitTime(self)




    def time(self):

        localctx = pcodeParser.TimeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 44, self.RULE_time)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 299
            self.timeexp()
            self.state = 301 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 300
                self.match(pcodeParser.WHITESPACE)
                self.state = 303 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==16):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class TimeexpContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def POSITIVE_FLOAT(self):
            return self.getToken(pcodeParser.POSITIVE_FLOAT, 0)

        def getRuleIndex(self):
            return pcodeParser.RULE_timeexp

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterTimeexp" ):
                listener.enterTimeexp(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitTimeexp" ):
                listener.exitTimeexp(self)




    def timeexp(self):

        localctx = pcodeParser.TimeexpContext(self, self._ctx, self.state)
        self.enterRule(localctx, 46, self.RULE_timeexp)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 305
            self.match(pcodeParser.POSITIVE_FLOAT)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class CommentContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def HASH(self):
            return self.getToken(pcodeParser.HASH, 0)

        def comment_text(self):
            return self.getTypedRuleContext(pcodeParser.Comment_textContext,0)


        def getRuleIndex(self):
            return pcodeParser.RULE_comment

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterComment" ):
                listener.enterComment(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitComment" ):
                listener.exitComment(self)




    def comment(self):

        localctx = pcodeParser.CommentContext(self, self._ctx, self.state)
        self.enterRule(localctx, 48, self.RULE_comment)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 307
            self.match(pcodeParser.HASH)
            self.state = 308
            self.comment_text()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Comment_textContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NEWLINE(self):
            return self.getToken(pcodeParser.NEWLINE, 0)

        def getRuleIndex(self):
            return pcodeParser.RULE_comment_text

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterComment_text" ):
                listener.enterComment_text(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitComment_text" ):
                listener.exitComment_text(self)




    def comment_text(self):

        localctx = pcodeParser.Comment_textContext(self, self._ctx, self.state)
        self.enterRule(localctx, 50, self.RULE_comment_text)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 313
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,43,self._ctx)
            while _alt!=1 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1+1:
                    self.state = 310
                    self.matchWildcard() 
                self.state = 315
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,43,self._ctx)

            self.state = 316
            _la = self._input.LA(1)
            if _la <= 0 or _la==25:
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class BlankContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def WHITESPACE(self, i:int=None):
            if i is None:
                return self.getTokens(pcodeParser.WHITESPACE)
            else:
                return self.getToken(pcodeParser.WHITESPACE, i)

        def getRuleIndex(self):
            return pcodeParser.RULE_blank

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBlank" ):
                listener.enterBlank(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBlank" ):
                listener.exitBlank(self)




    def blank(self):

        localctx = pcodeParser.BlankContext(self, self._ctx, self.state)
        self.enterRule(localctx, 52, self.RULE_blank)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 321
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,44,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    self.state = 318
                    self.match(pcodeParser.WHITESPACE) 
                self.state = 323
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,44,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class CommandContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def command_name(self):
            return self.getTypedRuleContext(pcodeParser.Command_nameContext,0)


        def time(self):
            return self.getTypedRuleContext(pcodeParser.TimeContext,0)


        def COLON(self):
            return self.getToken(pcodeParser.COLON, 0)

        def command_args(self):
            return self.getTypedRuleContext(pcodeParser.Command_argsContext,0)


        def WHITESPACE(self, i:int=None):
            if i is None:
                return self.getTokens(pcodeParser.WHITESPACE)
            else:
                return self.getToken(pcodeParser.WHITESPACE, i)

        def getRuleIndex(self):
            return pcodeParser.RULE_command

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCommand" ):
                listener.enterCommand(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCommand" ):
                listener.exitCommand(self)




    def command(self):

        localctx = pcodeParser.CommandContext(self, self._ctx, self.state)
        self.enterRule(localctx, 54, self.RULE_command)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 325
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==14:
                self.state = 324
                self.time()


            self.state = 327
            self.command_name()
            self.state = 336
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==23:
                self.state = 328
                self.match(pcodeParser.COLON)
                self.state = 332
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,46,self._ctx)
                while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                    if _alt==1:
                        self.state = 329
                        self.match(pcodeParser.WHITESPACE) 
                    self.state = 334
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,46,self._ctx)

                self.state = 335
                self.command_args()


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Command_nameContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def identifier(self):
            return self.getTypedRuleContext(pcodeParser.IdentifierContext,0)


        def getRuleIndex(self):
            return pcodeParser.RULE_command_name

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCommand_name" ):
                listener.enterCommand_name(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCommand_name" ):
                listener.exitCommand_name(self)




    def command_name(self):

        localctx = pcodeParser.Command_nameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 56, self.RULE_command_name)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 338
            self.identifier()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Command_argsContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NEWLINE(self):
            return self.getToken(pcodeParser.NEWLINE, 0)

        def HASH(self):
            return self.getToken(pcodeParser.HASH, 0)

        def getRuleIndex(self):
            return pcodeParser.RULE_command_args

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCommand_args" ):
                listener.enterCommand_args(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCommand_args" ):
                listener.exitCommand_args(self)




    def command_args(self):

        localctx = pcodeParser.Command_argsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 58, self.RULE_command_args)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 343
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,48,self._ctx)
            while _alt!=1 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1+1:
                    self.state = 340
                    self.matchWildcard() 
                self.state = 345
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,48,self._ctx)

            self.state = 346
            _la = self._input.LA(1)
            if _la <= 0 or _la==22 or _la==25:
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class IdentifierContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(pcodeParser.IDENTIFIER, 0)

        def getRuleIndex(self):
            return pcodeParser.RULE_identifier

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterIdentifier" ):
                listener.enterIdentifier(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitIdentifier" ):
                listener.exitIdentifier(self)




    def identifier(self):

        localctx = pcodeParser.IdentifierContext(self, self._ctx, self.state)
        self.enterRule(localctx, 60, self.RULE_identifier)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 348
            self.match(pcodeParser.IDENTIFIER)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Identifier_extContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(pcodeParser.IDENTIFIER, 0)

        def NEWLINE(self):
            return self.getToken(pcodeParser.NEWLINE, 0)

        def HASH(self):
            return self.getToken(pcodeParser.HASH, 0)

        def getRuleIndex(self):
            return pcodeParser.RULE_identifier_ext

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterIdentifier_ext" ):
                listener.enterIdentifier_ext(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitIdentifier_ext" ):
                listener.exitIdentifier_ext(self)




    def identifier_ext(self):

        localctx = pcodeParser.Identifier_extContext(self, self._ctx, self.state)
        self.enterRule(localctx, 62, self.RULE_identifier_ext)
        self._la = 0 # Token type
        try:
            self.state = 358
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,50,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 350
                self.match(pcodeParser.IDENTIFIER)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 354
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,49,self._ctx)
                while _alt!=1 and _alt!=ATN.INVALID_ALT_NUMBER:
                    if _alt==1+1:
                        self.state = 351
                        self.matchWildcard() 
                    self.state = 356
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,49,self._ctx)

                self.state = 357
                _la = self._input.LA(1)
                if _la <= 0 or _la==22 or _la==25:
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Inst_errorContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NEWLINE(self):
            return self.getToken(pcodeParser.NEWLINE, 0)

        def HASH(self):
            return self.getToken(pcodeParser.HASH, 0)

        def getRuleIndex(self):
            return pcodeParser.RULE_inst_error

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterInst_error" ):
                listener.enterInst_error(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitInst_error" ):
                listener.exitInst_error(self)




    def inst_error(self):

        localctx = pcodeParser.Inst_errorContext(self, self._ctx, self.state)
        self.enterRule(localctx, 64, self.RULE_inst_error)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 363
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,51,self._ctx)
            while _alt!=1 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1+1:
                    self.state = 360
                    self.matchWildcard() 
                self.state = 365
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,51,self._ctx)

            self.state = 366
            _la = self._input.LA(1)
            if _la <= 0 or _la==22 or _la==25:
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ErrorContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NEWLINE(self):
            return self.getToken(pcodeParser.NEWLINE, 0)

        def HASH(self):
            return self.getToken(pcodeParser.HASH, 0)

        def getRuleIndex(self):
            return pcodeParser.RULE_error

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterError" ):
                listener.enterError(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitError" ):
                listener.exitError(self)




    def error(self):

        localctx = pcodeParser.ErrorContext(self, self._ctx, self.state)
        self.enterRule(localctx, 66, self.RULE_error)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 371
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,52,self._ctx)
            while _alt!=1 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1+1:
                    self.state = 368
                    self.matchWildcard() 
                self.state = 373
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,52,self._ctx)

            self.state = 374
            _la = self._input.LA(1)
            if _la <= 0 or _la==22 or _la==25:
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








from __future__ import annotations
from typing import Any, Generic, TypeVar


CommandArgs = dict[str, Any]
""" Command arguments"""


class EngineCommand():
    """ Interface for commands runnable by engine. """
    def __init__(self, name) -> None:
        self.name: str = name
        self._cancelled: bool = False
        self._initialized: bool = False
        self._exec_started: bool = False
        self._exec_iterations: int = -1
        self._exec_complete: bool = False
        self._progress: bool | float = False
        self._finalized: bool = False

    def initialize(self):
        self._initialized = True

    def is_initialized(self) -> bool:
        return self._initialized

    def execute(self, args: CommandArgs) -> None:
        self._exec_started = True
        self._exec_iterations += 1

    def is_execution_started(self) -> bool:
        return self._exec_started

    def get_iteration_count(self) -> int:
        """ Returns number of iterations executed. The first iteration is number 0. """
        return self._exec_iterations

    def is_execution_complete(self) -> bool:
        return self._exec_complete

    def finalize(self):
        """ Overrides must call super().finalize() and must dispose the command. """
        self._finalized = True

    def is_finalized(self) -> bool:
        return self._finalized

    def cancel(self):
        self._cancelled = True

    def is_cancelled(self) -> bool:
        return self._cancelled

    def set_progress(self, progress: bool | float):
        self._progress = progress

    def get_progress(self) -> bool | float:
        return self._progress

    def set_complete(self):
        self._exec_complete = True

    def force(self):
        pass


TContext = TypeVar('TContext')


class ContextEngineCommand(Generic[TContext], EngineCommand):
    """ Extension of EngineCommand that provides a typed context.

    This allows Uod commands to extend EngineCommand without importing Engine.
    """
    def __init__(self, context: TContext, name: str) -> None:
        super().__init__(name)
        self.context: TContext = context



import json
import os
import subprocess

from openpectus.aggregator.aggregator_server import AggregatorServer

frontend_path = os.path.join(
    os.path.dirname(os.path.dirname(__file__)),
    "frontend",
)
openapi_json_path = os.path.join(
    frontend_path,
    "openapi.json",
)

if __name__ == "__main__":
    print(f"Generate/update openapi spec file {openapi_json_path}")
    with open(openapi_json_path, "wt") as f:
        json.dump(AggregatorServer().fastapi.openapi(), f, indent=2)

    cmd = "npm run generate"
    print(f"Generating typescript interfaces using command: {cmd}")
    retval = subprocess.check_call(cmd, cwd=frontend_path, shell=True)
    if retval == 0:
        print("Generation complete")
    else:
        print(f"Generation failed with error code: {retval}")



from enum import StrEnum
from typing import Literal
from openpectus.lang.exec.tags import SystemTagName, TagDirection

SystemTagName = SystemTagName
TagDirection = TagDirection

class EngineCommandEnum(StrEnum):
    START = "Start"
    STOP = "Stop"
    PAUSE = "Pause"
    UNPAUSE = "Unpause"
    HOLD = "Hold"
    UNHOLD = "Unhold"
    WAIT = "Wait"
    RESTART = "Restart"
    INFO = "Info"
    WARNING = "Warning"
    ERROR = "Error"

    @staticmethod
    def has_value(value: str):
        """ Determine if enum has this string value defined. Case sensitive. """
        return value in EngineCommandEnum.__members__.values()


class SystemStateEnum(StrEnum):
    Running = "Running"
    Paused = "Paused"
    Holding = "Holding"
    Waiting = "Waiting"
    Stopped = "Stopped"
    Restarting = "Restarting"


class MethodStatusEnum(StrEnum):
    OK = "OK"
    ERROR = "Error"


class ConnectionStatusEnum(StrEnum):
    """ Defines status for the connectedness of the Engine-Hardware connection. """
    Disconnected = "Disconnected"
    Connected = "Connected"


EntryDataType = Literal["str"] | Literal["int"] | Literal["float"] | Literal["auto"]



import logging
from logging.handlers import QueueHandler
from queue import Empty, SimpleQueue
import time

import openpectus.protocol.engine_messages as EM
import openpectus.protocol.models as Mdl
from openpectus.engine.engine import Engine
from openpectus.lang.exec.runlog import RunLogItem
from openpectus.lang.exec.tags import TagValue
from openpectus.lang.exec.uod import logger as uod_logger
from openpectus.engine.engine import frontend_logger as engine_logger
from openpectus.engine.archiver import logger as archiver_logger
from openpectus.engine.internal_commands_impl import logger as internal_cmds_logger

logger = logging.getLogger(__name__)

frontend_logging_queue: SimpleQueue[logging.LogRecord] = SimpleQueue()
frontend_logging_handler = QueueHandler(frontend_logging_queue)
frontend_logging_handler.setLevel(logging.WARN)

# add frontend error logging for selected loggers (this gives import cycle warning - could we reverse this instead?)
uod_logger.addHandler(frontend_logging_handler)
engine_logger.addHandler(frontend_logging_handler)
archiver_logger.addHandler(frontend_logging_handler)
internal_cmds_logger.addHandler(frontend_logging_handler)

MAX_SIZE_TagsUpdatedMsg = 100
""" The maximum number of tags to include in a single TagsUpdatedMsg message """

def to_model_tag(tag: TagValue) -> Mdl.TagValue:
    if tag.tick_time == 0.0:
        logger.warning(f"Tick time was 0 for tag {tag.name}")
        tag.tick_time = time.time()
    return Mdl.TagValue(
        name=tag.name,
        tick_time=tag.tick_time,
        value=tag.value,
        value_formatted=tag.value_formatted,
        value_unit=tag.unit,
        direction=tag.direction
    )

class EngineMessageBuilder():
    """ Collects data from engine and constructs engine messages """
    def __init__(self, engine: Engine) -> None:
        self.engine = engine

    def create_uod_info(self) -> EM.UodInfoMsg:
        return EM.UodInfoMsg(
            readings=[reading.as_reading_info() for reading in self.engine.uod.readings],
            commands=[command.as_command_info() for command in self.engine.uod.command_descriptions.values()],
            plot_configuration=self.engine.uod.plot_configuration,
            hardware_str=str(self.engine.uod.hwl),
            required_roles=self.engine.uod.required_roles,
            data_log_interval_seconds=self.engine.uod.data_log_interval_seconds)

    def collect_tag_updates(self, snapshot=False) -> list[Mdl.TagValue]:
        if snapshot:
            self.engine.notify_all_tags()
        tags: dict[str, Mdl.TagValue] = {}  # using dict to de-duplicate
        try:
            for _ in range(MAX_SIZE_TagsUpdatedMsg):
                tag_org = self.engine.tag_updates.get_nowait()
                tag = tag_org.as_readonly()
                tags[tag.name] = to_model_tag(tag)
                self.engine.tag_updates.task_done()
        except Empty:
            pass
        return [tag for tag in tags.values()]

    def create_tag_updates_snapshot_msg(self) -> EM.TagsUpdatedMsg:
        tags = self.collect_tag_updates(snapshot=True)
        return EM.TagsUpdatedMsg(tags=tags)

    def create_tag_updates_msg(self) -> EM.TagsUpdatedMsg | None:
        tags = self.collect_tag_updates()
        # print("tags with updates", str([t.name for t in tags]))
        if len(tags) > 0:
            return EM.TagsUpdatedMsg(tags=tags)

    def create_run_started_msg(self, run_id: str, tick_time: float) -> EM.RunStartedMsg:
        return EM.RunStartedMsg(run_id=run_id, started_tick=tick_time)

    def create_run_stopped_msg(self, run_id: str) -> EM.RunStoppedMsg:
        runlog_msg = self.create_runlog_msg(run_id)
        method = self.engine._method._method
        return EM.RunStoppedMsg(run_id=run_id, runlog=runlog_msg.runlog, method=method)

    def create_runlog_msg(self, run_id: str) -> EM.RunLogMsg:
        tag_names: list[str] = []
        for reading in self.engine.uod.readings:
            tag_names.append(reading.tag_name)

        def filter_value(tag_value: TagValue) -> bool:
            return tag_value.name in tag_names

        def to_line(item: RunLogItem) -> Mdl.RunLogLine:
            msg = Mdl.RunLogLine(
                id=item.id,
                command_name=item.name,
                start=item.start,
                end=item.end,
                progress=item.progress,
                start_values=[to_model_tag(t) for t in filter(filter_value, item.start_values)],
                end_values=[to_model_tag(t) for t in filter(filter_value, item.end_values)],
                forcible=item.forcible,
                cancellable=item.cancellable,
                forced=item.forced,
                cancelled=item.cancelled
            )
            return msg

        runlog = self.engine.runtimeinfo.get_runlog()
        return EM.RunLogMsg(
            id=runlog.id,
            run_id=run_id,
            runlog=Mdl.RunLog(lines=list(map(to_line, runlog.items)))
        )

    def create_error_log_msg(self) -> EM.ErrorLogMsg | None:
        log: Mdl.ErrorLog = Mdl.ErrorLog(entries=[])
        while not frontend_logging_queue.empty():
            log_entry = frontend_logging_queue.get_nowait()
            log.entries.append(Mdl.ErrorLogEntry(
                message=log_entry.getMessage(),
                created_time=log_entry.created,
                severity=log_entry.levelno))
        if len(log.entries) != 0:
            return EM.ErrorLogMsg(log=log)

    def create_control_state_msg(self) -> EM.ControlStateMsg:
        return EM.ControlStateMsg(control_state=Mdl.ControlState(
            is_running=self.engine._runstate_started,
            is_holding=self.engine._runstate_holding,
            is_paused=self.engine._runstate_paused
        ))

    def create_method_state_msg(self) -> EM.MethodStateMsg:
        state = self.engine.calculate_method_state()
        return EM.MethodStateMsg(method_state=state)



from __future__ import annotations
from enum import StrEnum
from typing import Any, Callable, Dict, List
from uuid import UUID


class InterpreterCommandEnum(StrEnum):
    """ Commands (instructions of type PCommand) that are executed by the interpreter """
    BASE = "Base"
    INCREMENT_RUN_COUNTER = "Increment run counter"
    RUN_COUNTER = "Run counter"

    @staticmethod
    def has_value(value: str):
        """ Determine if enum has this string value defined. Case sensitive. """
        return value in InterpreterCommandEnum.__members__.values()


# Represents part of Engine API


class CommandRequest():
    """ Represents a command request for engine to execute. """
    def __init__(self, name: str, source: str, exec_id: UUID | None = None, kvargs: dict[str, Any] = {}) -> None:
        self.name: str = name
        self.unparsed_args: str | None = None
        self.kvargs: dict[str, Any] | None = kvargs
        self.exec_id: UUID | None = exec_id
        self.source: str = source

        # allows tracking individual commands
        self.command_exec_id: UUID | None = None

    @staticmethod
    def from_user(name: str, unparsed_args: str | None = None) -> CommandRequest:
        cmd_req = CommandRequest(name=name, source='user')
        cmd_req.unparsed_args = unparsed_args
        return cmd_req

    @staticmethod
    def from_interpreter(name: str, exec_id: UUID | None, **kvargs) -> CommandRequest:
        if "unparsed_args" in kvargs.keys():
            cmd_req = CommandRequest(name=name, source="@interpreter", exec_id=exec_id)
            cmd_req.unparsed_args = kvargs["unparsed_args"]
            return cmd_req
        else:
            return CommandRequest(name=name, source="@interpreter", exec_id=exec_id, kvargs=kvargs)

    def __str__(self) -> str:
        return f"EngineCommand {self.name} | kvargs: {str(self.kvargs)} | unparsed_args: {self.unparsed_args}"

    def get_args(self) -> dict[str, Any]:
        if self.kvargs is None or (self.kvargs == {} and self.unparsed_args is not None):
            return {"unparsed_args": self.unparsed_args}
        else:
            return self.kvargs


# Represents command API towards interpreter
# Used by analyzers


class Command:
    """ Represents a named command. """
    def __init__(self, name: str, validatorFn: Callable[[str], bool] | None = None) -> None:
        self.name: str = name
        self.validatorFn = validatorFn

    def is_complete(self) -> bool:
        return True

    def validate_args(self, args: str) -> bool:
        valid = True
        if self.validatorFn is not None:
            valid = self.validatorFn(args)
        return valid

    def clone(self) -> Command:
        return Command(self.name)


class CommandCollection():
    """ Represents a case insensitive name/command dictionary. """

    def __init__(self) -> None:
        self.commands: Dict[str, Command] = {}

    @property
    def names(self) -> List[str]:
        """ Return the tag names in upper case. """
        return list(self.commands.keys())

    def __getitem__(self, tag_name: str):
        return self.commands[tag_name.upper()]

    def get(self, cmd_name: str) -> Command:
        if cmd_name is None or cmd_name.strip() == '':
            raise ValueError("cmd_name is None or empty")
        if not cmd_name.upper() in self.commands.keys():
            raise ValueError(f"Command name {cmd_name} not found")
        return self[cmd_name]

    def add(self, cmd: Command, exist_ok: bool = True):
        """ Add command to collection. If command name already exists and exist_ok is False, a ValueError is raised. """
        if cmd is None:
            raise ValueError("cmd is None")
        if cmd.name is None or cmd.name.strip() == '':
            raise ValueError("cmd name is None or empty")
        if cmd.name in self.commands.keys() and not exist_ok:
            raise ValueError(f"A command named {cmd.name} already exists")

        self.commands[cmd.name.upper()] = cmd

    def with_cmd(self, cmd: Command):
        self.add(cmd)
        return self

    def has(self, cmd_name: str) -> bool:
        if cmd_name is None or cmd_name.strip() == '':
            raise ValueError("cmd_name is None or empty")
        return cmd_name.upper() in self.commands.keys()

    def clone(self) -> CommandCollection:
        """ Returns a deep clone of the collection. """
        cmds = CommandCollection()
        for cmd in self.commands.values():
            cmds.add(cmd.clone())
        return cmds

    def merge_with(self, other: CommandCollection) -> CommandCollection:
        """ Returns a new CommandCollection with the combined commands of both collections.

        In case of duplicate commands names, tags from other collection are used.
        """
        cmds = CommandCollection()
        for cmd in self.commands.values():
            cmds.add(cmd)
        for cmd in other.commands.values():
            cmds.add(cmd)
        return cmds



import asyncio
import unittest
from multiprocessing import Process
from unittest import IsolatedAsyncioTestCase

import uvicorn
from fastapi import FastAPI
from fastapi_websocket_rpc import RpcChannel, WebsocketRPCEndpoint, RpcMethodsBase, WebSocketRpcClient
from fastapi_websocket_rpc.logger import get_logger
from openpectus.protocol.dispatch_interface import AGGREGATOR_RPC_WS_PATH

logger = get_logger("Test")

PORT = 7990
ws_url = f"ws://localhost:{PORT}/{AGGREGATOR_RPC_WS_PATH}"
trigger_url = f"http://localhost:{PORT}/trigger"

mock_engine_id = 'mock engine id'


async def on_delayed_client_connect(channel: RpcChannel):
    engine_id = await channel.other.get_engine_id_async()
    # self.assertEqual(engine_id, mock_engine_id)
    logger.info(engine_id)


async def on_client_connect(channel: RpcChannel):
    asyncio.create_task(on_delayed_client_connect(channel))


class RpcMethods(RpcMethodsBase):
    async def get_engine_id_async(self):
        return mock_engine_id


def setup_server_endpoint():
    server = FastAPI()
    # WebsockeRPCEndpoint has wrong types for its on_connect and on_disconnect.
    # It should be List[Callable[[RpcChannel], Awaitable[Any]]] instead of List[Coroutine]
    server_endpoint = WebsocketRPCEndpoint(on_connect=[on_client_connect])  # type: ignore
    server_endpoint.register_route(server, path=AGGREGATOR_RPC_WS_PATH)
    uvicorn.run(server, port=PORT)


class TestRpc(IsolatedAsyncioTestCase):
    async def test_get_engine_id(self):
        proc = Process(target=setup_server_endpoint, daemon=True)
        proc.start()
        rpc_url = f"ws://127.0.0.1:{PORT}{AGGREGATOR_RPC_WS_PATH}"
        rpc_methods = RpcMethods()
        rpc_client = WebSocketRpcClient(uri=rpc_url, methods=rpc_methods)
        await rpc_client.__aenter__()
        proc.terminate()


if __name__ == "__main__":
    unittest.main()



import logging
import os
from subprocess import Popen, STDOUT
import time
from typing import Any
import unittest
import tempfile

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


class ConsoleAppRunner():
    def __init__(self, program: str, args: list[str]) -> None:
        self.program = program
        tempfile.mktemp()
        self.logfile_name = tempfile.NamedTemporaryFile(suffix=".log").name
        logger.debug(f"Using temp file {self.logfile_name}")
        self.process: Popen[Any] | None = None
        self.args = args
        self.lines: list[str] = []

    def __enter__(self):
        return self

    def start(self):
        self.logfile = open(self.logfile_name, "wb")
        args = [self.program] + self.args
        self.process = Popen(args, stdout=self.logfile, stderr=STDOUT, encoding="utf-8")

    @property
    def output(self) -> str:
        for i in range(5):
            try:
                with open(self.logfile_name, "r") as reader:
                    return reader.read()
            except PermissionError:
                # This error occours when reading is performed while the file is being flushed to disk.
                # Just try again later.
                time.sleep(0.05)
        raise Exception("Unable to read {self.logfile_name}.")

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.process:
            logger.debug("Killing process")
            self.process.kill()
        try:
            self.logfile.close()
        except Exception:
            logger.error("Failed to close file")
        if os.path.exists(self.logfile_name):
            time.sleep(1)
            os.unlink(self.logfile_name)


def get_demo_uod_path() -> str:
    uod_file_path = os.path.join(
        os.path.dirname(  # openpectus
            os.path.dirname(  # openpectus/test
                os.path.dirname(__file__)  # openpectus/test/engine
            )
        ),
        'engine',  # openpectus/engine
        'configuration',  # openpectus/engine/configuration
        'demo_uod.py',  # openpectus/engine/configuration/demo_uod.py
    )
    return uod_file_path

class TestStartupConnection(unittest.TestCase):

    def assertHasOutput(self, runner: ConsoleAppRunner, expected_text: str, timeout_secs=5):
        output = runner.output
        if expected_text in output:
            return
        end_time = time.time() + timeout_secs
        while time.time() < end_time:
            output = runner.output
            if expected_text in output:
                return
            time.sleep(0.5)

        raise AssertionError(f"Expected text '{expected_text}' was not found in output",
                             f"Output is: {output}")

    def test_engine_can_connect_to_running_aggregator(self):
        port = "8727"
        with ConsoleAppRunner("pectus-aggregator", ["--port", port]) as aggregator, \
             ConsoleAppRunner("pectus-engine", ["--aggregator_port", port, "--uod", get_demo_uod_path()]) as engine:

            aggregator.start()
            self.assertHasOutput(aggregator, "Starting Open Pectus Aggregator")

            engine.start()
            self.assertHasOutput(engine, "Registering for engine id")

            self.assertHasOutput(aggregator, "Engine connected", 3)
            self.assertHasOutput(engine, "Changing state: Started -> Connected", 3)
            self.assertHasOutput(engine, "Starting engine on first steady-state", 3)


    def test_engine_can_connect_to_aggregator_started_after_engine(self):
        port = "8728"
        with ConsoleAppRunner("pectus-aggregator", ["--port", port]) as aggregator, \
             ConsoleAppRunner("pectus-engine", ["--aggregator_port", port, "--uod", get_demo_uod_path()]) as engine:

            engine.start()
            time.sleep(5)
            self.assertHasOutput(engine, "Registering for engine id")

            aggregator.start()
            time.sleep(5)
            self.assertHasOutput(aggregator, "Starting Open Pectus Aggregator")

            time.sleep(2)
            self.assertHasOutput(aggregator, "Engine connected")
            self.assertHasOutput(engine, "Changing state: CatchingUp -> Reconnected")
            self.assertHasOutput(engine, "Starting engine on first steady-state")




import unittest

from openpectus.lang.exec.tags import Tag, TagCollection
from openpectus.lang.exec.tags_impl import ReadingTag
from openpectus.lang.exec.uod import RegexNumber, RegexCategorical, UodBuilder, UodCommand


def create_minimal_builder() -> UodBuilder:
    builder = (
        UodBuilder()
        .with_instrument("TestUod")
        .with_author("Test Author", "test@openpectus.org")
        .with_filename(__file__)
        .with_hardware_none()
        .with_location("Test location")
    )
    return builder


class TestCommandDescriptions(unittest.TestCase):
    def test_units_filtered_by_regex(self):

        def exec_Area(cmd: UodCommand, number, number_unit):
            pass

        uod = (create_minimal_builder()
               .with_tag(ReadingTag("Area", unit="cm2"))
               .with_command_regex_arguments("Area", RegexNumber(units=["cm2", "dm2"]), exec_Area)
               ).build()

        uod.system_tags = TagCollection.create_system_tags()
        uod.validate_configuration()
        uod.build_commands()

        cmd = uod.command_descriptions["Area"]
        self.assertEqual(['cm2', "dm2"], cmd.argument_valid_units)

    def test_exec_fn_docstring(self):

        def exec_Area(cmd: UodCommand, number, number_unit):
            """ # Runs the Area command
            Area: 5 cm2
            """
            pass

        uod = (create_minimal_builder()
               .with_tag(ReadingTag("Area", unit="cm2"))
               .with_command_regex_arguments("Area", RegexNumber(units=["cm2", "dm2"]), exec_Area)
               ).build()

        uod.system_tags = TagCollection.create_system_tags()
        uod.validate_configuration()
        uod.build_commands()

        cmd = uod.command_descriptions["Area"]
        # Note removal of white space
        self.assertEqual(
            """# Runs the Area command\nArea: 5 cm2""",
            cmd.docstring
        )

    def test_get_docstring_pcode(self):
        def exec_Area(cmd: UodCommand, number, number_unit):
            """ # Runs the Area command
            Area: 5 cm2
            """
            pass

        uod = (create_minimal_builder()
               .with_tag(ReadingTag("Area", unit="cm2"))
               .with_command_regex_arguments("Area", RegexNumber(units=["cm2", "dm2"]), exec_Area)
               ).build()

        uod.system_tags = TagCollection.create_system_tags()
        uod.validate_configuration()
        uod.build_commands()

        cmd = uod.command_descriptions["Area"]
        pcode = cmd.get_docstring_pcode()
        self.assertEqual("# Runs the Area command\nArea: 5 cm2", pcode)

    def test_generate_example_pcode_number_with_unit(self):
        def exec_Area(cmd: UodCommand, number, number_unit):
            pass

        uod = (create_minimal_builder()
               .with_tag(ReadingTag("Area", unit="cm2"))
               .with_command_regex_arguments("Area", RegexNumber(units=["cm2", "dm2"]), exec_Area)
               ).build()

        uod.system_tags = TagCollection.create_system_tags()
        uod.validate_configuration()
        uod.build_commands()

        cmd = uod.command_descriptions["Area"]
        # remove comments
        pcode_examples = cmd.generate_pcode_examples()
        self.assertEqual(2, len(pcode_examples))
        self.assertEqual("Area: 0.5 cm2", pcode_examples[0])
        self.assertEqual("Area: 0.5 dm2", pcode_examples[1])

    def test_generate_example_pcode_w_unit(self):
        def exec_Vel(cmd: UodCommand, number, number_unit):
            pass

        uod = (create_minimal_builder()
               .with_tag(ReadingTag("Flow", unit="L/min"))
               .with_command_regex_arguments("Flow", RegexNumber(units=["L/h", "L/min"]), exec_Vel)
               ).build()

        uod.system_tags = TagCollection.create_system_tags()
        uod.validate_configuration()
        uod.build_commands()

        cmd = uod.command_descriptions["Flow"]
        pcode_examples = cmd.generate_pcode_examples()
        self.assertEqual(2, len(pcode_examples))
        self.assertEqual("Flow: 0.5 L/h", pcode_examples[0])
        self.assertEqual("Flow: 0.5 L/min", pcode_examples[1])

    def test_options_filtered_by_regex(self):

        def exec_Categorical(cmd: UodCommand, option: str):
            pass

        uod = (create_minimal_builder()
               .with_tag(Tag("Category", value=""))
               .with_command_regex_arguments("Categorical",
                                             RegexCategorical(exclusive_options=["A", "B"],
                                                              additive_options=["1", "2", "3"]),
                                             exec_Categorical)
               ).build()

        uod.system_tags = TagCollection.create_system_tags()
        uod.validate_configuration()
        uod.build_commands()

        cmd = uod.command_descriptions["Categorical"]
        self.assertEqual(["A", "B"], cmd.argument_exclusive_options)
        self.assertEqual(["1", "2", "3"], cmd.argument_additive_options)

    def test_generate_example_pcode_categorical(self):

        def exec_Categorical(cmd: UodCommand, option: str):
            pass

        uod = (create_minimal_builder()
               .with_tag(Tag("Category", value=""))
               .with_command_regex_arguments("Categorical",
                                             RegexCategorical(exclusive_options=["A", "B"],
                                                              additive_options=["1", "2", "3"]),
                                             exec_Categorical)
               ).build()

        uod.system_tags = TagCollection.create_system_tags()
        uod.validate_configuration()
        uod.build_commands()

        cmd = uod.command_descriptions["Categorical"]
        pcode_examples = cmd.generate_pcode_examples()
        self.assertEqual(9, len(pcode_examples))
        self.assertEqual("Categorical: A", pcode_examples[0])
        self.assertEqual("Categorical: B", pcode_examples[1])
        self.assertEqual("Categorical: 1", pcode_examples[2])
        self.assertEqual("Categorical: 2", pcode_examples[3])
        self.assertEqual("Categorical: 3", pcode_examples[4])
        self.assertEqual("Categorical: 1+2", pcode_examples[5])
        self.assertEqual("Categorical: 1+3", pcode_examples[6])
        self.assertEqual("Categorical: 2+3", pcode_examples[7])
        self.assertEqual("Categorical: 1+2+3", pcode_examples[8])
